from unittest.mock import patch

import pytest

from ballet.templating import start_new_feature
from ballet.util import work_in


@pytest.mark.parametrize(
    'branching',
    [True, False],
)
def test_start_new_feature(quickstart, caplog, branching):
    default_branch = quickstart.project.branch
    with work_in(quickstart.project.path):
        with patch('ballet.project.Project.from_cwd') as mock_from_cwd:
            mock_from_cwd.return_value = quickstart.project
            extra_context = {
                'username': 'username',
                'featurename': 'featurename',
            }
            start_new_feature(
                branching=branching, no_input=True, extra_context=extra_context
            )

    assert 'Start new feature successful' in caplog.text

    if branching:
        assert quickstart.project.branch == 'username/feature-featurename'
    else:
        assert quickstart.project.branch == default_branch



import importlib.machinery
import pathlib

from funcy import re_test

from ballet.project import relative_to_contrib
from ballet.util.fs import isemptyfile
from ballet.validation.base import BaseCheck

FEATURE_MODULE_NAME_REGEX = r'feature_(?P<featurename>[a-zA-Z0-9_]+)\.py'
SUBPACKAGE_NAME_REGEX = r'user_(?P<username>[a-zA-Z0-9_]+)'


class ProjectStructureCheck(BaseCheck):
    """Base class for implementing new Feature API checks

    Args:
        project (Project): project
    """

    def __init__(self, project):
        self.project = project


class IsAdditionCheck(ProjectStructureCheck):

    def check(self, diff):
        """Check that the diff represents the addition of a new file"""
        assert diff.change_type == 'A'


class IsPythonSourceCheck(ProjectStructureCheck):

    def check(self, diff):
        """Check that the new file introduced is a python source file"""
        path = diff.b_path
        assert any(
            path.endswith(ext)
            for ext in importlib.machinery.SOURCE_SUFFIXES
        )


class WithinContribCheck(ProjectStructureCheck):

    def check(self, diff):
        """Check that the new file is within the contrib subdirectory"""
        path = diff.b_path
        contrib_path = self.project.config.get('contrib.module_path')
        assert pathlib.Path(contrib_path) in pathlib.Path(path).parents


class SubpackageNameCheck(ProjectStructureCheck):

    def check(self, diff):
        """Check that the name of the subpackage within contrib is valid

        The package name must match ``user_[a-zA-Z0-9_]+``.
        """
        relative_path = relative_to_contrib(diff, self.project)
        subpackage_name = relative_path.parts[0]
        assert re_test(SUBPACKAGE_NAME_REGEX, subpackage_name)


class RelativeNameDepthCheck(ProjectStructureCheck):

    def check(self, diff):
        """Check that the new file introduced is at the proper depth

        The proper depth is 2 (contrib/user_example/new_file.py)
        """
        relative_path = relative_to_contrib(diff, self.project)
        assert len(relative_path.parts) == 2


class ModuleNameCheck(ProjectStructureCheck):

    def check(self, diff):
        r"""Check that the new file introduced has a valid name

        The module can either be an __init__.py file or must
        match ``feature_[a-zA-Z0-9_]+\.\w+``.
        """
        filename = pathlib.Path(diff.b_path).parts[-1]
        is_valid_feature_module_name = re_test(
            FEATURE_MODULE_NAME_REGEX, filename)
        is_valid_init_module_name = filename == '__init__.py'
        assert is_valid_feature_module_name or is_valid_init_module_name


class IfInitModuleThenIsEmptyCheck(ProjectStructureCheck):

    def check(self, diff):
        """Check that if the new file is __init__.py, then it is empty"""
        path = pathlib.Path(diff.b_path)
        filename = path.parts[-1]
        if filename == '__init__.py':
            abspath = self.project.path.joinpath(path)
            assert isemptyfile(abspath)



import pytest

from ballet.util.code import blacken_code, get_source, is_valid_python


def test_is_valid_python():
    code = '1'
    result = is_valid_python(code)
    assert result


def test_is_valid_python_invalid():
    code = 'this is not valid python code'
    result = is_valid_python(code)
    assert not result


def test_blacken_code():
    input = '''\
    x = {  'a':37,'b':42,

    'c':927}
    '''.strip()

    expected = 'x = {"a": 37, "b": 42, "c": 927}'.strip()
    actual = blacken_code(input).strip()

    assert actual == expected


def test_blacken_code_nothing_changed():
    input = '1\n'
    expected = '1\n'
    actual = blacken_code(input)

    assert actual == expected


@pytest.mark.xfail
def test_get_source():
    get_source(None)



__all__ = []
try:
    from skits.feature_extraction import AutoregressiveTransformer
    __all__.append('AutoregressiveTransformer')

    from skits.feature_extraction import SeasonalTransformer
    __all__.append('SeasonalTransformer')

    from skits.feature_extraction import IntegratedTransformer
    __all__.append('IntegratedTransformer')

    from skits.feature_extraction import RollingMeanTransformer
    __all__.append('RollingMeanTransformer')

    from skits.feature_extraction import TrendTransformer
    __all__.append('TrendTransformer')

    from skits.feature_extraction import FourierTransformer
    __all__.append('FourierTransformer')

    from skits.preprocessing import ReversibleImputer
    __all__.append('ReversibleImputer')

    from skits.preprocessing import DifferenceTransformer
    __all__.append('DifferenceTransformer')

    from skits.preprocessing import LogTransformer
    __all__.append('LogTransformer')

    from skits.preprocessing import HorizonTransformer
    __all__.append('HorizonTransformer')
except ImportError:
    pass



import os
from unittest.mock import ANY, mock_open, patch

import numpy as np
import pandas as pd
import pytest

from ballet.util.io import (
    _check_ext, _write_tabular_h5, _write_tabular_pickle, write_tabular,)


@pytest.fixture
def array():
    return np.arange(10).reshape(2, 5)


@pytest.fixture
def frame():
    return pd.util.testing.makeDataFrame()


def test_check_ext_valid():
    ext = '.py'
    expected = ext
    _check_ext(ext, expected)


def test_check_ext_invalid_throws():
    ext = '.py'
    expected = '.txt'
    with pytest.raises(ValueError):
        _check_ext(ext, expected)


@patch('ballet.util.io._write_tabular_pickle')
@patch('ballet.util.io._write_tabular_h5')
def test_write_tabular(mock_write_tabular_h5,
                       mock_write_tabular_pickle):
    obj = object()
    filepath = '/foo/bar/baz.h5'
    write_tabular(obj, filepath)
    mock_write_tabular_h5.assert_called_once_with(obj, filepath)

    obj = object()
    filepath = '/foo/bar/baz.pkl'
    write_tabular(obj, filepath)
    mock_write_tabular_pickle.assert_called_once_with(obj, filepath)

    obj = object()
    filepath = '/foo/bar/baz.xyz'
    with pytest.raises(NotImplementedError):
        write_tabular(obj, filepath)


@patch('builtins.open', new_callable=mock_open)
@patch('pickle.dump')
def test_write_tabular_pickle_ndarray(mock_dump, mock_open, array):
    obj = array
    filepath = '/foo/bar/baz.pkl'
    _write_tabular_pickle(obj, filepath)
    mock_dump.assert_called_with(obj, ANY)


@patch('builtins.open', new_callable=mock_open)
def test_write_tabular_pickle_ndframe(mock_open, frame):
    obj = frame
    filepath = '/foo/bar/baz.pkl'

    with patch.object(obj, 'to_pickle') as mock_to_pickle:
        _write_tabular_pickle(obj, filepath)

    mock_to_pickle.assert_called_with(filepath)


def test_write_tabular_pickle_nonarray_raises():
    obj = object()
    filepath = '/foo/bar/baz.pkl'
    with pytest.raises(NotImplementedError):
        _write_tabular_pickle(obj, filepath)


def test_write_tabular_h5_ndarray(tmp_path, array):
    obj = array
    filepath = tmp_path.joinpath('baz.h5')
    _write_tabular_h5(obj, filepath)

    file_size = os.path.getsize(filepath)
    assert file_size > 0


def test_write_tabular_h5_ndframe(frame):
    obj = frame
    filepath = '/foo/bar/baz.h5'

    with patch.object(obj, 'to_hdf') as mock_to_hdf:
        _write_tabular_h5(obj, filepath)

    mock_to_hdf.assert_called_with(filepath, key=ANY)


@pytest.mark.xfail
def test_read_tabular():
    raise NotImplementedError


@pytest.mark.xfail
def test_save_model():
    raise NotImplementedError


@pytest.mark.xfail
def test_save_predictions():
    raise NotImplementedError



import pathlib
import re
from typing import Iterable, Iterator, Optional, Tuple

import git
import requests
from funcy import collecting, complement, lfilter, re_find, silent
from github import Github
from github.Repository import Repository
from stacklog import stacklog

from ballet.exc import BalletError
from ballet.util import one_or_raise
from ballet.util.log import logger

FILE_CHANGES_COMMIT_RANGE = '{a}...{b}'
REV_REGEX = r'[a-zA-Z0-9_/^@{}-]+'
COMMIT_RANGE_REGEX = re.compile(
    fr'(?P<a>{REV_REGEX})\.\.(?P<thirddot>\.?)(?P<b>{REV_REGEX})')
GIT_PUSH_FAILURE = (
    git.PushInfo.REJECTED |
    git.PushInfo.REMOTE_REJECTED |
    git.PushInfo.REMOTE_FAILURE |
    git.PushInfo.ERROR
)
DEFAULT_BRANCH = 'master'


class Differ:

    def diff(self) -> git.DiffIndex:
        a, b = self._get_diff_endpoints()
        return a.diff(b)

    def _get_diff_endpoints(self) -> Tuple[git.Diffable, git.Diffable]:
        raise NotImplementedError


class CustomDiffer(Differ):

    def __init__(self, endpoints: Tuple[git.Diffable, git.Diffable]):
        self.endpoints = endpoints

    def _get_diff_endpoints(self) -> Tuple[git.Diffable, git.Diffable]:
        return self.endpoints


class PullRequestBuildDiffer(Differ):
    """Diff files from this pull request against a comparison ref

    Args:
        repo: repo
    """

    def __init__(self, repo: git.Repo):
        self.repo = repo
        self._check_environment()

    def _check_environment(self):
        raise NotImplementedError


class NoOpDiffer(PullRequestBuildDiffer):
    """A differ that returns an empty changeset"""

    def diff(self):
        return []

    def _check_environment(self):
        pass


def can_use_local_differ(repo: Optional[git.Repo]):
    """On some non-master branch"""
    return repo is not None and repo.head.ref.name != 'master'


class LocalPullRequestBuildDiffer(PullRequestBuildDiffer):

    @property
    def _ref_name(self) -> str:
        return self.repo.head.ref.name

    @property
    def _ref_path(self) -> str:
        return self.repo.head.ref.path

    def _check_environment(self):
        assert self._ref_name != 'master'

    def _get_diff_endpoints(self) -> Tuple[git.Diffable, git.Diffable]:
        a = self.repo.rev_parse('master')
        b = self.repo.rev_parse(self._ref_name)
        return a, b


def can_use_local_merge_differ(repo: Optional[git.Repo]):
    """Check the repo HEAD is on master after a merge commit

    Checks for two qualities of the current project:

    1. The project repo's head is the master branch
    2. The project repo's head commit is a merge commit.

    Note that fast-forward style merges will not cause the second condition
    to evaluate to true.
    """
    if repo:
        return (
            get_branch(repo) == 'master' and
            is_merge_commit(repo.head.commit)
        )


class LocalMergeBuildDiffer(Differ):
    """Diff files on a merge commit on the current active branch.

    Merge parent order is guaranteed such that parent 1 is HEAD and parent 2
    is topic [1].

    Attributes:
        repo: The repository to check the merge diff on. Must be currently on
            a branch where the most recent commit is a merge.

    References:
        [1] https://git-scm.com/book/en/v2/Git-Tools-Advanced-Merging
    """

    def __init__(self, repo: git.Repo):
        self.repo = repo
        self._check_environment()

    def _check_environment(self):
        assert is_merge_commit(self.repo.head.commit)

    def _get_diff_endpoints(self) -> Tuple[git.Diffable, git.Diffable]:
        parents = self.repo.head.commit.parents
        return parents[0], parents[1]


def make_commit_range(a: str, b: str) -> str:
    return FILE_CHANGES_COMMIT_RANGE.format(a=a, b=b)


def get_diff_endpoints_from_commit_range(
    repo: git.Repo, commit_range: str
) -> Tuple[git.Diffable, git.Diffable]:
    """Get endpoints of a diff given a commit range

    The resulting endpoints can be diffed directly::

        a, b = get_diff_endpoints_from_commit_range(repo, commit_range)
        a.diff(b)

    For details on specifying git diffs, see ``git diff --help``.
    For details on specifying revisions, see ``git help revisions``.

    Args:
        repo: Repo object initialized with project root
        commit_range: commit range as would be interpreted by ``git
            diff`` command. Unfortunately only patterns of the form ``a..b``
            and ``a...b`` are accepted. Note that the latter pattern finds the
            merge-base of a and b and uses it as the starting point for the
            diff.

    Returns:
        starting commit, ending commit (inclusive)

    Raises:
        ValueError: commit_range is empty or ill-formed

    See also:

        <https://stackoverflow.com/q/7251477>
    """
    if not commit_range:
        raise ValueError('commit_range cannot be empty')

    result = re_find(COMMIT_RANGE_REGEX, commit_range)
    if not result:
        raise ValueError(
            f'Expected diff str of the form \'a..b\' or \'a...b\' '
            f'(got {commit_range})')
    a, b = result['a'], result['b']
    a, b = repo.rev_parse(a), repo.rev_parse(b)
    if result['thirddot']:
        a = one_or_raise(repo.merge_base(a, b))
    return a, b


def get_repo(repo: Optional[git.Repo] = None) -> git.Repo:
    if repo is None:
        repo = git.Repo(pathlib.Path.cwd(),
                        search_parent_directories=True)
    return repo


@silent
def get_branch(repo: Optional[git.Repo] = None) -> str:
    repo = get_repo(repo)
    branch = repo.head.ref.name
    return branch


def switch_to_new_branch(repo: git.Repo, name: str):
    new_branch = repo.create_head(name)
    repo.head.ref = new_branch


def is_merge_commit(commit: git.Commit) -> bool:
    return len(commit.parents) > 1


def set_config_variables(repo: git.Repo, variables: dict):
    """Set config variables

    Args:
        repo: repo
        variables: entries of the form ``'user.email': 'you@example.com'``
    """
    with repo.config_writer() as writer:
        for k, value in variables.items():
            section, option = k.split('.')
            writer.set_value(section, option, value)
        writer.release()


def get_pull_requests(owner: str, repo: str, state: str = 'closed') -> dict:
    base = 'https://api.github.com'
    q = f'/repos/{owner}/{repo}/pulls'
    url = base + q
    headers = {
        'Accept': 'application/vnd.github.v3+json'
    }
    params = {
        'state': state,
        'base': 'master',
        'sort': 'created',
        'direction': 'asc',
    }
    res = requests.get(url, headers=headers, params=params)
    res.raise_for_status()
    return res.json()


@collecting
def get_pull_request_outcomes(owner: str, repo: str) -> Iterator[str]:
    prs = get_pull_requests(owner, repo, state='closed')
    for pr in prs:
        if pr['merged_at'] is not None:
            yield 'accepted'
        else:
            yield 'rejected'


def did_git_push_succeed(push_info: git.remote.PushInfo) -> bool:
    """Check whether a git push succeeded

    A git push succeeded if it was not "rejected" or "remote rejected",
    and if there was not a "remote failure" or an "error".

    Args:
        push_info: push info
    """
    return push_info.flags & GIT_PUSH_FAILURE == 0


def create_github_repo(github: Github, owner: str, name: str) -> Repository:
    """Create the repo ``:owner/:name``

    The authenticated account must have the permissions to create the desired
    repo.

    1. if the desired owner is the user, then this is straightforward
    2. if the desired owner is an organization, then the user must have
       permission to create a new repo for the organization

    Returns:
        the created repository

    Raises:
        github.GithubException.BadCredentialsException: if the token does not
            have permission to create the desired repo
    """
    user = github.get_user()
    if user.login == owner:
        return user.create_repo(name)
    else:
        return github.get_organization(owner).create_repo(name)


@stacklog(logger.info, 'Pushing branches to remote')
def push_branches_to_remote(
    repo: git.Repo,
    remote_name: str,
    branches: Iterable[str]
):
    """Push selected branches to origin

    Similar to::

        $ git push origin branch1:branch1 branch2:branch2

    Raises:
        ballet.exc.BalletError: Push failed in some way
    """
    remote = repo.remote(remote_name)
    result = remote.push([
        f'{b}:{b}'
        for b in branches
    ])
    failures = lfilter(complement(did_git_push_succeed), result)
    if failures:
        for push_info in failures:
            logger.error(
                f'Failed to push ref {push_info.local_ref.name} to '
                f'{push_info.remote_ref.name}')
        raise BalletError('Push failed')



from unittest.mock import create_autospec, patch

import git
import pytest
from github import Github

from ballet.util.git import (
    create_github_repo, did_git_push_succeed, get_pull_request_outcomes,
    get_pull_requests, make_commit_range, push_branches_to_remote,)


def test_make_commit_range():
    a = 'abc1234'
    b = 'def4321'
    expected_commit_range = 'abc1234...def4321'
    actual_commit_range = make_commit_range(a, b)
    assert actual_commit_range == expected_commit_range


@pytest.mark.xfail
def test_get_diff_endpoints_from_commit_range():
    raise NotImplementedError


@pytest.mark.xfail
def test_get_repo():
    raise NotImplementedError


@pytest.mark.xfail
def test_get_branch():
    raise NotImplementedError


@pytest.mark.xfail
def test_switch_to_new_branch():
    raise NotImplementedError


@pytest.mark.xfail
def test_set_config_variables():
    raise NotImplementedError


@patch('requests.get')
def test_get_pull_requests(mock_requests_get):
    owner = 'foo'
    repo = 'bar'
    state = 'closed'
    get_pull_requests(owner, repo, state=state)

    (url, ), kwargs = mock_requests_get.call_args
    assert owner in url
    assert repo in url
    assert 'headers' in kwargs
    assert 'params' in kwargs
    assert kwargs['params']['state'] == state


@patch('ballet.util.git.get_pull_requests')
def test_get_pull_request_outcomes(mock_get_pull_requests):
    mock_get_pull_requests.return_value = [
        {
            'id': 1,
            "created_at": "2011-01-26T19:01:12Z",
            "updated_at": "2011-01-26T19:01:12Z",
            "closed_at": "2011-01-26T19:01:12Z",
            "merged_at": "2011-01-26T19:01:12Z",
        },
        {
            "created_at": "2011-01-26T19:03:19Z",
            "updated_at": "2011-01-26T19:03:19Z",
            "closed_at": "2011-01-26T19:04:01Z",
            "merged_at": None,
        }
    ]
    owner = 'foo'
    repo = 'bar'

    expected = ['accepted', 'rejected']
    actual = get_pull_request_outcomes(owner, repo)
    assert actual == expected
    mock_get_pull_requests.assert_called_once_with(
        owner, repo, state='closed')


def test_did_git_push_succeed():
    local_ref = None
    remote_ref_string = None
    remote = None

    flags = 0
    push_info = git.remote.PushInfo(flags, local_ref, remote_ref_string,
                                    remote)
    assert did_git_push_succeed(push_info)

    flags = git.remote.PushInfo.ERROR
    push_info = git.remote.PushInfo(flags, local_ref, remote_ref_string,
                                    remote)
    assert not did_git_push_succeed(push_info)


@pytest.fixture
def github():
    g = create_autospec(Github)
    g.get_user.return_value.login = 'octocat'
    return g


@pytest.mark.parametrize(
    'owner',
    ['octocat', 'github-dot-com'],
)
def test_create_github_repo(github, owner):
    name = 'Hello-World'
    create_github_repo(github, owner, name)

    if owner == 'octocat':
        create_repo = github.get_user.return_value.create_repo
    else:
        create_repo = github.get_organization.return_value.create_repo

    create_repo.assert_called_once_with(name)


@patch('git.Repo.remote')
def test_push_branches_to_remote(mock_remote, mock_repo):
    mock_push = mock_remote.return_value.push
    remote_name = 'origin'
    branch_name = 'master'
    branches = [branch_name]
    push_branches_to_remote(mock_repo, remote_name, branches)
    mock_push.assert_called_once_with([f'{branch_name}:{branch_name}'])



"""
If the extra is not installed, then the test passes. If the extra is
installed, we should be able to import the last attribute exported from that
module.
"""


def test_category_encoders():
    try:
        import category_encoders  # noqa F401
    except ImportError:
        pass
    else:
        from ballet.eng.external.category_encoders import (  # noqa F401
            WOEEncoder,)


def test_feature_engine():
    try:
        import feature_engine  # noqa F401
    except ImportError:
        pass
    else:
        from ballet.eng.external.feature_engine import (  # noqa f401
            YeoJohnsonTransformer,)


def test_featuretools():
    try:
        import featuretools  # noqa F401
    except ImportError:
        pass
    else:
        from ballet.eng.external.featuretools import (  # noqa F401
            DFSTransformer,)


def test_skits():
    try:
        import skits  # noqa F401
    except ImportError:
        pass
    else:
        from ballet.eng.external.skits import HorizonTransformer  # noqa F401


def test_sklearn():
    from ballet.eng.external.sklearn import KNNImputer  # noqa F401


def test_tsfresh():
    try:
        import tsfresh  # noqa F401
    except ImportError:
        pass
    else:
        from ballet.eng.external.tsfresh import FeatureAugmenter  # noqa F401



from ballet.project import FeatureEngineeringProject

import {{ cookiecutter.package_slug }} as package
from {{ cookiecutter.package_slug }}.features.encoder import get_target_encoder
from {{ cookiecutter.package_slug }}.load_data import load_data


api = FeatureEngineeringProject(
    package=package,
    encoder=get_target_encoder(),
    load_data=load_data,
)



from ballet.validation.common import load_spec
from ballet.validation.feature_acceptance.validator import (
    AlwaysAccepter, RandomAccepter,)
from ballet.validation.project_structure.validator import (
    ProjectStructureValidator,)


def test_load_spec_from_name(caplog):
    spec = 'ballet.validation.project_structure.validator.ProjectStructureValidator'  # noqa
    expected_class = ProjectStructureValidator
    cls, params = load_spec(spec)
    assert cls is expected_class
    assert isinstance(params, dict)


def test_load_spec_noparams():
    spec = {
        'name': 'ballet.validation.feature_acceptance.validator.AlwaysAccepter',  # noqa
    }
    expected_class = AlwaysAccepter
    cls, params = load_spec(spec)
    assert cls is expected_class


def test_load_spec_params():
    threshold = 0.88
    spec = {
        'name': 'ballet.validation.feature_acceptance.validator.RandomAccepter',  # noqa
        'params': {
            'threshold': threshold
        }
    }
    expected_class = RandomAccepter
    cls, params = load_spec(spec)
    assert cls is expected_class
    assert params['threshold'] == threshold



from unittest.mock import patch

import numpy as np
import pandas as pd
import pytest

from ballet.eng.misc import IdentityTransformer
from ballet.feature import Feature
from ballet.validation.feature_acceptance.validator import (
    CompoundAccepter, GFSSFAccepter, MutualInformationAccepter, NeverAccepter,
    RandomAccepter, VarianceThresholdAccepter,)
from tests.util import load_regression_data


def test_noop_accepter():
    X_df = y_df = y = None
    existing_features = []
    feature = None

    expected = False

    accepter = NeverAccepter(X_df, y_df, X_df, y, existing_features, feature)
    actual = accepter.judge()

    assert expected == actual


@patch('random.uniform', return_value=0.0)  # makes sure feature is accepted
def test_random_accepter(mock_uniform):
    X_df = y_df = y = None
    existing_features = []
    candidate_feature = None

    expected = True

    accepter = RandomAccepter(
        X_df, y_df, X_df, y, existing_features, candidate_feature)
    actual = accepter.judge()

    assert expected == actual


@pytest.fixture
def sample_data():
    X_df, y_df = load_regression_data(
        n_informative=1, n_uninformative=14)
    y = y_df
    return X_df, y_df, y


def test_gfssf_accepter_init(sample_data):
    X_df, y_df, y = sample_data

    feature_1 = Feature(
        input='A_0',
        transformer=IdentityTransformer(),
        source='1st Feature')
    feature_2 = Feature(
        input='Z_0',
        transformer=IdentityTransformer(),
        source='2nd Feature')

    features = [feature_1]
    candidate_feature = feature_2

    accepter = GFSSFAccepter(
        X_df, y_df, X_df, y, features, candidate_feature)

    assert accepter is not None


@patch('numpy.var', return_value=0.0)
def test_variance_threshold_accepter(mock_var, sample_data):
    expected = False
    X_df, y_df, y = sample_data
    feature = Feature(
        input='A_0',
        transformer=IdentityTransformer(),
        source='1st Feature')
    accepter = VarianceThresholdAccepter(
        X_df, y_df, X_df, y, [], feature)
    actual = accepter.judge()

    assert expected == actual


def test_variance_threshold_accepter_feature_group():
    expected = True
    # variance is 0.25 per column, > 0.05 threshold
    X = pd.DataFrame(np.eye(2))
    y = None
    feature = Feature(
        input=[0, 1],
        transformer=IdentityTransformer(),
        source='1st Feature')
    accepter = VarianceThresholdAccepter(
        X, y, X, y, [], feature)
    actual = accepter.judge()

    assert expected == actual


@patch(
    'ballet.validation.feature_acceptance.validator.estimate_mutual_information',  # noqa
    return_value=0.99
)
def test_mutual_information_accepter(_, sample_data):
    expected = True
    X_df, y_df, y = sample_data
    feature = Feature(
        input='A_0',
        transformer=IdentityTransformer(),
        source='1st Feature')
    accepter = MutualInformationAccepter(
        X_df, y_df, X_df, y, [], feature)
    actual = accepter.judge()

    assert expected == actual


@pytest.mark.parametrize(
    'handle_nan_targets, expected',
    [
        ('fail', False),
        ('ignore', True)
    ],
)
def test_mutual_information_accepter_nans(handle_nan_targets, expected):
    X_df = pd.DataFrame({'A': [1, 2, 3]})
    y = np.array([np.nan, 2, 3]).reshape(-1, 1)
    feature = Feature(
        input='A',
        transformer=IdentityTransformer())
    accepter = MutualInformationAccepter(
        X_df, y, X_df, y, [], feature, handle_nan_targets=handle_nan_targets)
    actual = accepter.judge()
    assert expected == actual


def test_compound_accepter(sample_data):
    expected = False
    X_df, y_df, y = sample_data
    agg = 'all'
    specs = [
        'ballet.validation.feature_acceptance.validator.AlwaysAccepter',
        {
            'name': 'ballet.validation.feature_acceptance.validator.RandomAccepter',  # noqa
            'params': {
                'p': 0.00,
            }
        }
    ]
    feature = Feature(
        input='A_0',
        transformer=IdentityTransformer(),
        source='1st Feature')
    accepter = CompoundAccepter(
        X_df, y_df, X_df, y, [], feature, agg=agg, specs=specs
    )
    actual = accepter.judge()

    assert expected == actual



import traceback
from collections import Counter
from copy import deepcopy
from inspect import signature
from typing import (
    Callable, Collection, List, NamedTuple, Optional, Sequence, Tuple, Union,
    cast,)

import numpy as np
import pandas as pd
from funcy import identity, is_seqcont, select_values
from sklearn.base import BaseEstimator
from sklearn.preprocessing import FunctionTransformer
from sklearn_pandas.pipeline import TransformerPipeline

from ballet.eng import BaseTransformer, IdentityTransformer, SubsetTransformer
from ballet.exc import UnsuccessfulInputConversionError
from ballet.util import DeepcopyMixin, asarray2d, indent, quiet
from ballet.util.log import TRACE, logger
from ballet.util.typing import OneOrMore, TransformerLike

RobustTransformer = Union[TransformerPipeline, 'DelegatingRobustTransformer']


def make_robust_transformer(
    transformer: OneOrMore[TransformerLike]
) -> RobustTransformer:
    """Convert to robust transformer or pipeline

    Convert to either a single DelegatingRobustTransformer or a
    TransformerPipeline where each transformer in the pipeline is a
    DelegatingRobustTransformer.
    """
    if is_seqcont(transformer):
        transformer = cast(Collection[TransformerLike], transformer)
        transformers = list(
            map(desugar_transformer, transformer))
        for t in transformers:
            _validate_transformer_api(t)
        return make_robust_transformer_pipeline(transformers)
    else:
        transformer = cast(TransformerLike, transformer)
        transformer = desugar_transformer(transformer)
        _validate_transformer_api(transformer)
        return DelegatingRobustTransformer(transformer)


def _name_estimators(
    estimators: Sequence[BaseEstimator]
) -> List[Tuple[str, BaseEstimator]]:
    """Generate names for estimators.

    Adapted from sklearn.pipeline._name_estimators
    """

    def get_name(estimator):
        if isinstance(estimator, DelegatingRobustTransformer):
            return get_name(estimator._transformer)

        return type(estimator).__name__.lower()

    names = list(map(get_name, estimators))
    counter = dict(Counter(names))
    counter = select_values(lambda x: x > 1, counter)

    for i in reversed(range(len(estimators))):
        name = names[i]
        if name in counter:
            names[i] += "-%d" % counter[name]
            counter[name] -= 1

    return list(zip(names, estimators))


def make_transformer_pipeline(
    steps: Sequence[BaseTransformer],
) -> TransformerPipeline:
    """Construct a TransformerPipeline from the given estimators.

    Source: sklearn_pandas.cont_method
    """
    return TransformerPipeline(_name_estimators(steps))


def make_robust_transformer_pipeline(
    steps: Collection[BaseTransformer]
) -> TransformerPipeline:
    """Construct a transformer pipeline of DelegatingRobustTransformers"""
    return make_transformer_pipeline([
        DelegatingRobustTransformer(step) for step in steps
    ])


class ConversionApproach(NamedTuple):
    name: str
    convert: Callable
    caught: Collection[type]

    isokay: Optional[Callable[[Exception], bool]] = lambda exc: False
    """Opportunity to catch other exceptions that match a condition"""


def catch_bare_exception_sanitize_array(e):
    """See pandas-dev/pandas#35744"""
    return isinstance(e, Exception) and 'Data must be 1-dimensional' in str(e)


class DelegatingRobustTransformer(DeepcopyMixin, BaseTransformer):
    """Robust transformer that delegates to underlying transformer

    This transformer is robust against different typed and shaped input data.
    It tries a variety of input data conversion approaches and passes the
    result to the underlying transformer, using the first approach that works.

    Args:
        transformer: a transformer object with fit and transform methods

    Raises:
        UnsuccessfulInputConversionError: If none of the conversion approaches
            work.

    """

    DEFAULT_CAUGHT = (ValueError, TypeError)

    CONVERSION_APPROACHES = [
        ConversionApproach('identity', identity, DEFAULT_CAUGHT),
        ConversionApproach(
            'series', pd.Series, DEFAULT_CAUGHT,
            isokay=catch_bare_exception_sanitize_array),  # FIXME
        ConversionApproach('dataframe', pd.DataFrame, DEFAULT_CAUGHT),
        ConversionApproach('array', np.asarray, DEFAULT_CAUGHT),
        ConversionApproach('asarray2d', asarray2d, ()),
    ]

    def __init__(self, transformer: BaseTransformer):
        self._transformer = transformer
        self._stored_conversion_approach = None

    def get_params(self, deep=False):
        transformer = self.__getattribute__('_transformer')
        if deep:
            transformer = deepcopy(transformer)
        return {
            'transformer': transformer
        }

    def __getattr__(self, attr):
        if '_transformer' in self.__dict__:
            return getattr(self._transformer, attr)
        else:
            raise AttributeError

    def __getstate__(self):
        return self.__dict__.copy()

    def __setstate__(self, state):
        self.__dict__.update(state)

    def __repr__(self):
        name = type(self).__name__
        return f'{name}({self._transformer!r})'

    def __str__(self):
        return str(self._transformer)

    @property
    def _tname(self) -> str:
        return type(self._transformer).__name__

    def fit(self, X, y=None, **kwargs):
        # don't return the result of transformer.fit because it is the
        # underlying transformer, not this robust transformer
        self._call_robust(self._transformer.fit, X, y, kwargs)

        # instead, return this robust transformer
        return self

    def transform(self, X, y=None, **kwargs):
        return self._call_robust(self._transformer.transform, X, y, kwargs)

    @staticmethod
    def _call_with_convert(method, convert, X, y, kwargs):
        if y is not None:
            return method(convert(X), y=convert(y), **kwargs)
        else:
            return method(convert(X), **kwargs)

    @quiet
    def _call_robust(self, method, X, y, kwargs):
        if self._stored_conversion_approach is not None:
            approach = self._stored_conversion_approach
            self._log_attempt_using_stored_approach(approach)
            convert = approach.convert
            try:
                result = self._call_with_convert(method, convert, X, y, kwargs)
                self._log_success_using_stored_approach(approach)
                return result
            except Exception as e:
                self._log_failure_using_stored_approach(approach, e)
                raise
        else:
            for approach in DelegatingRobustTransformer.CONVERSION_APPROACHES:
                try:
                    self._log_attempt(approach)
                    result = self._call_with_convert(
                        method, approach.convert, X, y, kwargs)
                    self._log_success(approach)
                    self._stored_conversion_approach = approach
                    return result
                except approach.caught as e:
                    self._log_catch(approach, e)
                    continue
                except Exception as e:
                    if approach.isokay(e):
                        self._log_catch(approach, e)
                        continue
                    else:
                        self._log_error(approach, e)
                        raise

            self._log_failure_no_more_approaches()
            raise UnsuccessfulInputConversionError

    def _log_attempt_using_stored_approach(self, approach):
        logger.log(
            TRACE,
            f'{self._tname}: '
            f'Attempting to convert using stored, '
            f'previously-successful approach {approach.name!r}')

    def _log_failure_using_stored_approach(self, approach, e):
        pretty_tb = self._get_pretty_tb()
        exc_name = type(e).__name__
        logger.debug(
            f'{self._tname}: '
            f'Conversion unexpectedly failed using stored, '
            f'previously-successful approach {approach.name!r} '
            f'because of error {exc_name!r}\n\n{pretty_tb}')

    def _log_success_using_stored_approach(self, approach):
        logger.log(
            TRACE,
            f'{self._tname}: '
            f'Conversion with stored, previously-successful approach '
            f'{approach.name!r} succeeded!')

    def _log_attempt(self, approach):
        logger.log(
            TRACE,
            f'{self._tname}: '
            f'Attempting to convert using approach {approach.name!r}...')

    def _get_pretty_tb(self):
        tb = traceback.format_exc()
        pretty_tb = indent(tb, n=8)
        return pretty_tb

    def _log_catch(self, approach, e):
        pretty_tb = self._get_pretty_tb()
        exc_name = type(e).__name__
        logger.debug(
            f'{self._tname}: '
            f'Conversion approach {approach.name!r} didn\'t work so we\'ll '
            f'try another approach, '
            f'caught exception {exc_name!r}\n\n{pretty_tb}')

    def _log_error(self, approach, e):
        pretty_tb = self._get_pretty_tb()
        exc_name = type(e).__name__
        logger.debug(
            f'{self._tname}: '
            f'Conversion failed during {approach.name!r} because of '
            f'an unrecoverable error {exc_name!r}\n\n{pretty_tb}')

    def _log_success(self, approach):
        logger.log(
            TRACE,
            f'{self._tname}: '
            f'Conversion approach {approach.name!r} succeeded!')

    def _log_failure_no_more_approaches(self):
        logger.debug('Conversion failed, and we\'re not sure why...')


def _validate_transformer_api(transformer: BaseTransformer):
    if not all(
        hasattr(transformer, attr)
        for attr in ['fit', 'transform', 'fit_transform']
    ):
        raise ValueError('Transformer object missing required attribute')

    sig_fit = signature(transformer.fit)
    if '(X, y=None' not in str(sig_fit):
        raise ValueError(f'Invalid signature for transformer.fit: {sig_fit}')

    sig_transform = signature(transformer.transform)
    if '(X' not in str(sig_transform):
        raise ValueError(
            f'Invalid signature for transformer.transform: {sig_transform}')


def desugar_transformer(
    transformer: TransformerLike,
) -> BaseTransformer:
    """Replace transformer syntactic sugar with actual transformer

    The following syntactic sugar is supported:
    - `None` is replaced with an IdentityTransformer
    - a callable (function or lambda) is replaced with a FunctionTransformer
        that wraps that callable
    - a tuple (input, transformer) is replaced with a SubsetTransformer
    """
    if transformer is None:
        return IdentityTransformer()
    elif callable(transformer) and not isinstance(transformer, type):
        return FunctionTransformer(transformer)
    elif isinstance(transformer, tuple):
        return SubsetTransformer(*transformer)
    else:
        transformer = cast(BaseTransformer, transformer)
        return transformer


def get_transformer_primitives(
    transformer: TransformerLike
) -> List[str]:
    """Get the primitives used in this transformer or pipeline

    The primitives are just the class names underlying the transformer or
    pipeline.
    """
    if isinstance(transformer, DelegatingRobustTransformer):
        return [transformer._tname]
    else:
        _transformer = cast(TransformerPipeline, transformer)
        return [
            t._tname
            for _, t in _transformer.steps
        ]



import os
import os.path
import pathlib
from shutil import copyfile
from typing import Callable, Iterator, List, Tuple

from funcy import partial, suppress

from ballet.exc import BalletError
from ballet.util.log import logger
from ballet.util.typing import Pathy


def spliceext(filepath: Pathy, s: str) -> str:
    """Add s into filepath before the extension"""
    root, ext = os.path.splitext(filepath)
    return root + s + ext


def replaceext(filepath: Pathy, new_ext: str) -> str:
    """Replace any existing file extension with a new one

    If the new extension is the empty string, all existing extensions will
    be removed.

    Example::

        >>> replaceext('/foo/bar.txt', 'py')
        '/foo/bar.py'
        >>> replaceext('/foo/bar.txt', '.doc')
        '/foo/bar.doc'

    Args:
        filepath: file path
        new_ext: new file extension; if a leading dot is not included, it will
            be added.
    """
    if new_ext and not new_ext.startswith('.'):
        new_ext = '.' + new_ext
    return str(pathlib.Path(filepath).with_suffix(new_ext))


def splitext2(filepath: Pathy) -> Tuple[str, str, str]:
    """Split filepath into root, filename, ext"""
    root, filename = os.path.split(filepath)
    filename, ext = os.path.splitext(filename)
    return root, filename, ext


def isemptyfile(filepath: Pathy) -> bool:
    """Determine if the file both exists and isempty"""
    exists = os.path.exists(filepath)
    if exists:
        filesize = os.path.getsize(filepath)
        return filesize == 0
    else:
        return False


def synctree(
    src: Pathy,
    dst: Pathy,
    onexist: Callable[[pathlib.Path], None] = None
) -> List[Tuple[pathlib.Path, str]]:
    """Recursively sync files at directory src to dst

    This is more or less equivalent to::

       cp -n -R ${src}/ ${dst}/

    If a file at the same path exists in src and dst, it is NOT overwritten
    in dst. Pass ``onexist`` in order to raise an error on such conditions.

    Args:
        src: source directory
        dst: destination directory, does not need to exist
        onexist: function to call if file exists at destination,
            takes the full path to destination file as only argument

    Returns:
        changes made by synctree, list of tuples of the form
        ("/absolute/path/to/file", "<kind>") where the change kind is one of
        "dir" (new directory was created) or "file" (new file was created).
    """
    src = pathlib.Path(src).resolve()
    dst = pathlib.Path(dst).resolve()

    if not src.is_dir():
        raise ValueError

    if dst.exists() and not dst.is_dir():
        raise ValueError

    if onexist is None:
        def _onexist(path): pass
        onexist = _onexist

    return _synctree(src, dst, onexist)


def _synctree(
    src: pathlib.Path,
    dst: pathlib.Path,
    onexist: Callable[[pathlib.Path], None]
) -> List[Tuple[pathlib.Path, str]]:
    result = []
    cleanup = []
    try:
        for _root, dirnames, filenames in os.walk(src):
            root = pathlib.Path(_root)
            relative_dir = root.relative_to(src)

            for dirname in dirnames:
                dstdir = dst.joinpath(relative_dir, dirname)
                if dstdir.exists():
                    if not dstdir.is_dir():
                        raise BalletError
                else:
                    logger.debug(f'Making directory: {dstdir!s}')
                    dstdir.mkdir()
                    result.append((dstdir, 'dir'))
                    cleanup.append(partial(os.rmdir, dstdir))

            for filename in filenames:
                srcfile = root.joinpath(filename)
                dstfile = dst.joinpath(relative_dir, filename)
                if dstfile.exists():
                    onexist(dstfile)
                else:
                    logger.debug(f'Copying file to destination: {dstfile!s}')
                    copyfile(srcfile, dstfile)
                    result.append((dstfile, 'file'))
                    cleanup.append(partial(os.unlink, dstfile))

    except Exception:
        with suppress(Exception):
            for f in reversed(cleanup):
                f()
        raise

    return result


def pwalk(d: Pathy, **kwargs) -> Iterator[pathlib.Path]:
    """Similar to os.walk but with pathlib.Path objects"""
    for _dirpath, dirnames, filenames in os.walk(d, **kwargs):
        dirpath = pathlib.Path(_dirpath)
        for p in dirnames + filenames:
            yield dirpath.joinpath(p)



import copy
import sys

import numpy as np
import pandas as pd
import pytest

from ballet.util import (
    DeepcopyMixin, asarray2d, dfilter, falsy, get_arr_desc, has_nans, indent,
    load_sklearn_df, make_plural_suffix, nonnegative, quiet, truthy,)
from ballet.util.log import logger
from ballet.util.testing import assert_array_equal


def test_asarray2d_shape_n_x_2():
    # case: second dimension is > 1
    a = np.zeros((3, 2))
    result = asarray2d(a)
    assert_array_equal(result, a)


def test_asarray2d_shape_n_x_1():
    # case: second dimension == 1
    a = np.zeros((3, 1))
    result = asarray2d(a)
    assert_array_equal(result, a)


def test_asarray2d_shape_n():
    # case: second dimension not present
    a = np.zeros((3,))
    result = asarray2d(a)
    expected_shape = (3, 1)
    assert result.shape == expected_shape
    assert_array_equal(np.ravel(result), a)


def test_asarray2d_series():
    # case: pd.Series
    a = np.zeros((3,))
    ser = pd.Series(a)
    result = asarray2d(ser)
    assert result.shape[1] >= 1
    assert_array_equal(
        result, asarray2d(a)
    )


def test_asarray2d_df():
    # case: pd.DataFrame
    a = np.zeros((3, 2))
    df = pd.DataFrame(a)
    result = asarray2d(df)
    assert result.shape == df.shape
    assert result.shape[1] >= 1
    assert_array_equal(result, a)


def test_get_arr_desc_array():
    a = np.ones((2, 2))
    expected = 'ndarray (2, 2)'
    actual = get_arr_desc(a)
    assert actual == expected


def test_get_arr_desc_frame():
    df = pd.DataFrame()
    expected = 'DataFrame (0, 0)'
    actual = get_arr_desc(df)
    assert actual == expected


def test_get_arr_desc_object():
    obj = object()
    expected = 'object <no shape>'
    actual = get_arr_desc(obj)
    assert actual == expected


def test_indent():
    text = (
        'Hello\n'
        '  world\n'
        '...hi'
    )
    expected = (
        '    Hello\n'
        '      world\n'
        '    ...hi'
    )
    actual = indent(text, n=4)
    assert actual == expected


def test_make_plural_suffix_plural():
    objs = ['honda', 'toyota']
    suffix = make_plural_suffix(objs)
    actual = f'car{suffix}'
    expected = 'cars'
    assert actual == expected


def test_make_plural_suffix_singular():
    objs = ['honda']
    suffix = make_plural_suffix(objs)
    actual = f'car{suffix}'
    expected = 'car'
    assert actual == expected


def test_has_nans():
    objs_with_nans = [
        pd.DataFrame(data={'x': [1, np.nan], 'y': [np.nan, 2]}),
        pd.DataFrame(data={'x': [1, np.nan]}),
        pd.Series(data=[1, np.nan]),
        np.array([[1, np.nan], [np.nan, 2]]),
        np.array([[1, np.nan]]),
        np.array([1, np.nan]),
        np.array([1, np.nan]).T,
        np.array(np.nan),
    ]

    objs_without_nans = [
        pd.DataFrame(data={'x': [1, 0], 'y': [0, 2]}),
        pd.DataFrame(data={'x': [1, 0]}),
        pd.Series(data=[1, 0]),
        np.array([[1, 0], [0, 2]]),
        np.array([[1, 0]]),
        np.array([1, 0]),
        np.array([1, 0]).T,
        np.array(0),
    ]

    for obj in objs_with_nans:
        assert has_nans(obj)

    for obj in objs_without_nans:
        assert not has_nans(obj)


def test_dfilter():
    @dfilter(lambda x: x >= 0)
    def numbers():
        return [-1, 2, 0, -2]

    actual = numbers()
    expected = [2, 0]
    assert actual == expected


def test_load_sklearn_df():
    name = 'iris'
    X_df, y_df = load_sklearn_df(name)

    # validation on X_df
    assert X_df is not None
    assert isinstance(X_df, pd.DataFrame)

    # validation on y_df
    assert y_df is not None
    assert isinstance(y_df, pd.Series)


def test_quiet_stderr():
    @quiet
    def f():
        print('bar', file=sys.stderr)

    # doesn't check that nothing is printed to sys.stderr, but makes sure
    # the stream is reset properly (esp on py<3.5)
    stderr = sys.stderr
    f()
    assert sys.stderr is stderr


def test_deepcopy_mixin():
    class E(Exception):
        pass

    class A:
        def __init__(self, a):
            self.a = a

        def __deepcopy__(self, memo):
            raise E

    class B(DeepcopyMixin, A):
        pass

    a = A(1)
    with pytest.raises(E):
        copy.deepcopy(a)

    b = B(1)
    copy.deepcopy(b)


def test_falsy():
    matrix = (
        (False, True),
        (True, False),
        ('false', True),  # i.e., is falsy
        ('', True),
        ('true', False),  # i.e, is not falsy
        ('123', False),
        (73, False),
    )
    for input, expected in matrix:
        actual = falsy(input)
        assert expected == actual

        # truthy is complement of falsy
        actual = truthy(input)
        assert expected != actual


def test_nonnegative_positive_output():
    @nonnegative()
    def func():
        return 1

    assert 1 == func()


def test_nonnegative_negative_output():
    @nonnegative(name="Result")
    def func():
        return -1

    assert 0 == func()


def test_nonnegative_negative_introspection(caplog):
    @nonnegative()
    def estimate_something():
        return -1

    with caplog.at_level('WARNING', logger=logger.name):
        estimate_something()

    assert "Something" in caplog.text



from unittest.mock import patch

import numpy as np
import pytest

from ballet.util import asarray2d
from ballet.util.testing import assert_array_equal
from ballet.validation.entropy import (
    NEIGHBORS_ALGORITHM, NEIGHBORS_METRIC, _compute_empirical_probability,
    _compute_epsilon, _compute_n_points_within_radius_i,
    _compute_volume_unit_ball, _estimate_cont_entropy, _estimate_disc_entropy,
    _is_column_cont, _is_column_disc, _make_neighbors,
    estimate_conditional_information, estimate_entropy,
    estimate_mutual_information,)


def test_make_neighbors():
    nn = _make_neighbors()
    assert NEIGHBORS_ALGORITHM == nn.algorithm
    assert NEIGHBORS_METRIC == nn.metric


def test_compute_nx_i():
    # Note the chebyshev distance is used
    n = 5
    x = np.array([
        [0, 0],
        [1, 0],
        [0, 1],
        [1, 1],
        [0.5, 0.5],
    ])
    radius = np.array([
        [0.7],
        [10],
        [1],
        [0.7],
        [0.1],
    ])
    expected_nx = np.array([
        2,  # 0,0 and 0.5, 0.5
        5,  # all points
        2,  # would be all points, but should specifically exclude
            # points on the margin
        2,  # 1, 1 and 0.5, 0.5
        1   # just 0.5, 0.5
    ])

    nn = _make_neighbors().fit(x)
    for i in range(n):
        x_i = x[i:i + 1, :]  # require a (1,m) row array
        radius_i = radius[i]
        nx_i = _compute_n_points_within_radius_i(nn, x_i, radius_i)
        expected_nx_i = expected_nx[i]
        assert expected_nx_i == nx_i


def test_compute_empirical_probability():
    x = [1, 1, 2, 3, 2, 1, 1, 2]
    expected_pk = np.array([4 / 8, 3 / 8, 1 / 8])
    expected_events = np.array([[1], [2], [3]])
    pk, events = _compute_empirical_probability(x)

    assert_array_equal(expected_pk, pk)
    assert_array_equal(expected_events, events)


def test_compute_volume_unit_ball_chebyshev():
    metric = 'chebyshev'
    expected_volume = 1
    for d in [1, 2, 5, 11]:
        volume = _compute_volume_unit_ball(d, metric=metric)
        assert expected_volume == volume


def test_compute_volume_unit_ball_euclidean():
    metric = 'euclidean'
    volume_upper_bound = 1
    for d in [1, 2, 5, 11]:
        volume = _compute_volume_unit_ball(d, metric=metric)
        assert volume <= volume_upper_bound


def test_compute_epsilon():
    # data looks like this:
    # |         x
    # |       x
    # |     x
    # |   x
    # | x
    # |---------|
    #
    x = np.array([
        [0.5, 0.5],
        [1.5, 1.5],
        [2.5, 2.5],
        [3.5, 3.5],
        [4.5, 4.5],
    ])

    # note k is 3 and distance is chebyshev
    expected_epsilon = np.array([
        [2 * 3.0],
        [2 * 2.0],
        [2 * 2.0],  # has two neighbors at distance 2
        [2 * 2.0],
        [2 * 3.0],
    ])

    epsilon = _compute_epsilon(x)

    assert_array_equal(expected_epsilon, epsilon)


def test_disc_entropy_constant_vals_1d():
    """If x (column vector) is constant, then H(x) = 0"""
    same_val_arr_ones = np.ones((50, 1))
    H_hat = _estimate_disc_entropy(same_val_arr_ones)
    assert 0 == H_hat


def test_disc_entropy_constant_vals_2d():
    """If each column in x (matrix), then H(x) = 0"""
    same_val_arr_zero = np.zeros((50, 1))
    same_val_arr_ones = np.ones((50, 1))
    same_multi_val_arr = np.concatenate(
        (same_val_arr_ones, same_val_arr_zero), axis=1)
    H_hat = _estimate_disc_entropy(same_multi_val_arr)
    assert 0 == H_hat


def test_disc_entropy_two_values():
    """Entropy of fair coin ~= log(2)"""
    same_val_arr_zero = np.zeros((50, 1))
    same_val_arr_ones = np.ones((50, 1))
    diff_val_arr = np.concatenate(
        (same_val_arr_ones, same_val_arr_zero), axis=0)

    expected_h = np.log(2)
    H_hat = _estimate_disc_entropy(diff_val_arr)
    assert round(abs(expected_h - H_hat), 7) == 0


def test_is_column_disc():
    x = asarray2d(np.arange(50))
    result = _is_column_disc(x)
    assert result


def test_is_column_cont():
    x = asarray2d(np.random.rand(50))
    result = _is_column_cont(x)
    assert result


@pytest.mark.skip(reason='skipping')
@patch('ballet.validation.entropy._get_disc_columns')
def test_cont_disc_entropy_differs_disc(get_disc_columns):
    """Expect cont, disc columns to have different entropy"""
    disc = asarray2d(np.arange(50))

    # we run into trouble here because as disc as *actually* discrete,
    # epsilon would not be calculated (it is set to some dummy value of
    # -inf). instead, we patch get_disc_columns and "force" epsilon to be
    # calculated
    epsilon = _compute_epsilon(disc)
    H_cont = _estimate_cont_entropy(disc, epsilon)

    H_disc = _estimate_disc_entropy(disc)

    assert H_cont != H_disc


def test_cont_disc_entropy_differs_cont():
    """Expect cont, disc columns to have different entropy"""
    cont = asarray2d(np.arange(50)) + 0.5
    epsilon = _compute_epsilon(cont)

    H_cont = _estimate_cont_entropy(cont, epsilon)
    H_disc = _estimate_disc_entropy(cont)

    assert H_cont != H_disc


def test_entropy_multiple_disc():
    same_val_arr_zero = np.zeros((50, 1))
    same_val_arr_ones = np.ones((50, 1))
    # The 0.5 forces float => classified as continuous
    cont_val_arange = asarray2d(np.arange(50) + 0.5)
    all_disc_arr = np.concatenate(
        (same_val_arr_ones, same_val_arr_zero), axis=1)
    mixed_val_arr = np.concatenate(
        (all_disc_arr, cont_val_arange), axis=1)

    all_disc_h = estimate_entropy(all_disc_arr)
    mixed_h = estimate_entropy(mixed_val_arr)
    assert mixed_h > all_disc_h, \
        'Expected adding continuous column increases entropy'


def test_mi_uninformative():
    x = np.reshape(np.arange(1, 101), (-1, 1))
    y = np.ones((100, 1))
    mi = estimate_mutual_information(x, y)
    h_z = estimate_entropy(x)
    assert h_z / 4 > mi, \
        'uninformative column should have no information'


def test_mi_informative():
    x = np.reshape(np.arange(1, 101), (-1, 1))
    y = np.reshape(np.arange(1, 101), (-1, 1))
    mi = estimate_mutual_information(x, y)
    h_y = estimate_entropy(y)
    assert mi > h_y / 4, \
        'exact copy columns should have high information'


def test_cmi_high_info_uninformative_z():
    # redundant copies have little information
    x = np.reshape(np.arange(1, 101), (-1, 1))
    y = np.reshape(np.arange(1, 101), (-1, 1))

    # exact copies of y should have lots of information
    useless_z = np.ones((100, 1))
    cmi = estimate_conditional_information(x, y, useless_z)
    mi = estimate_mutual_information(x, y)
    assert round(abs(cmi - mi)) == 0, \
        'uninformative z should not affect mutual information score'


def test_cmi_redundant_info():
    x = np.reshape(np.arange(1, 101), (-1, 1))
    y = np.reshape(np.arange(1, 101), (-1, 1))
    exact_z = np.reshape(np.arange(1, 101), (-1, 1))

    h_y = estimate_entropy(y)
    cmi = estimate_conditional_information(x, y, exact_z)
    assert h_y / 4 > cmi, \
        'redundant copies should have little information'



from os import PathLike
from types import ModuleType
from typing import Optional, Union

import pandas as pd
from funcy import cached_property

from ballet.discovery import discover as _discover
from ballet.feature import Feature
from ballet.project import FeatureEngineeringProject, Project
from ballet.validation.common import subsample_data_for_validation
from ballet.validation.feature_acceptance import validate_feature_acceptance
from ballet.validation.feature_api import validate_feature_api
from ballet.validation.main import _load_validator_class_params


class Client:
    """User client for validating features

    Provides a simple interface to validating a given feature.

    Args:
        prj: a way to create a ballet ``Project``, either with an
            already-created object, the project's top-level module, or a path
            within the project. If none of these are provided, will attempt to
            detect the project by ascending from the current working directory.
    """

    def __init__(
        self,
        prj: Union[Project, ModuleType, str, PathLike, None] = None
    ):
        self._prj = prj

    @cached_property
    def project(self) -> Project:
        """Access ballet-specific project info"""
        if self._prj is None:
            return Project.from_cwd()
        else:
            if isinstance(self._prj, Project):
                return self._prj
            elif isinstance(self._prj, ModuleType):
                return Project(self._prj)
            elif isinstance(self._prj, (str, PathLike)):
                return Project.from_path(self._prj)
            else:
                raise ValueError('not supported')

    @property
    def api(self) -> FeatureEngineeringProject:
        """Access feature engineering API of this project"""
        return self.project.api

    def _load_validation_data(
        self,
        X_df: pd.DataFrame,
        y_df: pd.DataFrame,
        subsample: bool
    ):
        if X_df is None or y_df is None:
            _X_df, _y_df = self.api.load_data()
        if X_df is None:
            X_df = _X_df
        if y_df is None:
            y_df = _y_df
        if subsample:
            X_df, y_df = subsample_data_for_validation(X_df, y_df)
        return X_df, y_df

    def validate_feature_api(
        self,
        feature: Feature,
        X_df: Optional[pd.DataFrame] = None,
        y_df: Optional[pd.DataFrame] = None,
        subsample: bool = False,
    ) -> bool:
        """Check that this feature satisfies the expected feature API"""
        X_df, y_df = self._load_validation_data(X_df, y_df, subsample)
        result = self.api.engineer_features(X_df, y_df)
        return validate_feature_api(
            feature, result.X_df, result.y, False, log_advice=True)

    def validate_feature_acceptance(
        self,
        feature: Feature,
        X_df: Optional[pd.DataFrame] = None,
        y_df: Optional[pd.DataFrame] = None,
        subsample: bool = False
    ) -> bool:
        """Evaluate the performance of this feature on the development data"""
        X_df, y_df = self._load_validation_data(X_df, y_df, subsample)
        result = self.api.engineer_features(X_df, y_df)
        accepter_class = _load_validator_class_params(
            self.project, 'validation.feature_accepter')
        return validate_feature_acceptance(
            accepter_class, feature, result.features, result.X_df,
            result.y_df, result.X_df, result.y, False)

    def discover(self, input=None, primitive=None) -> pd.DataFrame:
        features = self.api.features
        encoder = self.api.encoder

        # recover if we can't load data
        try:
            X_df, y_df = self.api.load_data()
            y = encoder.fit_transform(y_df)
        except Exception:
            X_df = y_df = y = None

        return _discover(
            features, X_df, y_df, y, input=input, primitive=primitive)

    discover.__doc__ = _discover.__doc__


b = Client()



import pathlib
import shutil
import tempfile
from typing import NamedTuple
from unittest.mock import patch

import git
import numpy as np
import pandas as pd
import pytest
import responses as _responses

import ballet
from ballet.project import Project
from ballet.templating import render_project_template
from ballet.util import work_in

from .util import set_ci_git_config_variables, tree


@pytest.fixture
def tempdir():
    """Tempdir fixture using tempfile.TemporaryDirectory"""
    with tempfile.TemporaryDirectory() as d:
        yield pathlib.Path(d)


@pytest.fixture
def testdatadir():
    return pathlib.Path(__file__).resolve().parent.joinpath('testdata')


def _mock_repo(tempdir):
    repo = git.Repo.init(str(tempdir))
    set_ci_git_config_variables(repo)
    return repo


@pytest.fixture
def mock_repo(tempdir):
    """Create a new repo"""
    yield _mock_repo(tempdir)


class QuickstartResult(NamedTuple):
    project: Project
    tempdir: pathlib.Path
    project_slug: str
    package_slug: str
    repo: git.Repo


@pytest.fixture
def quickstart(tempdir):
    """
    $ cd tempdir
    $ ballet quickstart
    $ tree .
    """
    # cd tempdir
    with work_in(tempdir):

        project_slug = 'foo-bar'
        package_slug = 'foo_bar'
        extra_context = {
            'project_name': project_slug.capitalize(),
            'project_slug': project_slug,
            'package_slug': package_slug,
        }

        # ballet quickstart
        render_project_template(no_input=True,
                                extra_context=extra_context,
                                output_dir=tempdir)

        # tree .
        tree(tempdir)

        project = Project.from_path(tempdir.joinpath(project_slug))
        repo = project.repo

        yield QuickstartResult(
            project, tempdir, project_slug, package_slug, repo
        )


@pytest.fixture
def project_template_copy(tempdir):
    old_path = ballet.templating.PROJECT_TEMPLATE_PATH
    new_path = tempdir.joinpath('templates', 'project_template')
    shutil.copytree(old_path, new_path)

    with patch('ballet.templating.PROJECT_TEMPLATE_PATH', new_path):
        tree(new_path)
        yield new_path


@pytest.fixture
def responses():
    with _responses.RequestsMock() as rsps:
        yield rsps


class SampleData(NamedTuple):
    df: pd.DataFrame
    X: pd.DataFrame
    y: pd.DataFrame


@pytest.fixture
def sample_data():
    df = pd.DataFrame(
        data={
            'country': ['USA', 'USA', 'Canada', 'Japan'],
            'year': [2001, 2002, 2001, 2002],
            'size': [np.nan, -11, 12, 0.0],
            'strength': [18, 110, np.nan, 101],
            'happy': [False, True, False, False]
        }
    ).set_index(['country', 'year'])
    X = df[['size', 'strength']]
    y = df[['happy']]
    return SampleData(df, X, y)



import logging
from unittest.mock import Mock, patch

import pytest

import ballet.util.log
from ballet.exc import SkippedValidationTest
from ballet.validation.main import (  # noqa F401
    _check_project_structure, _evaluate_feature_performance,
    _load_validator_class_params, _prune_existing_features,
    _validate_feature_api, validate, validation_stage,)
from ballet.validation.project_structure.validator import (
    ProjectStructureValidator,)


def test_validation_stage_success(caplog):
    caplog.set_level(logging.DEBUG, logger=ballet.util.log.logger.name)

    @validation_stage('do something')
    def call():
        return True

    call()

    assert 'DONE' in caplog.text


def test_validation_stage_skipped(caplog):
    caplog.set_level(logging.DEBUG, logger=ballet.util.log.logger.name)

    @validation_stage('do something')
    def call():
        raise SkippedValidationTest

    call()

    assert 'SKIPPED' in caplog.text


def test_validation_stage_failure():
    exc = ValueError

    @validation_stage('do something')
    def call():
        raise exc

    with pytest.raises(exc):
        call()


@patch('ballet.validation.common.load_spec')
def test_load_validator_class_params_noparams(mock_load_spec, caplog):
    """Check that _load_validator_class_params works"""

    mock_load_spec.return_value = (ProjectStructureValidator, {})

    mock_project = Mock()
    modname = 'ballet.validation.project_structure.validator'
    clsname = 'ProjectStructureValidator'
    path = modname + '.' + clsname
    mock_project.config.get.return_value = path
    config_key = None

    with caplog.at_level(logging.DEBUG, logger=ballet.util.log.logger.name):
        make_validator = _load_validator_class_params(
            mock_project, config_key)
    assert clsname in caplog.text

    validator = make_validator(mock_project)
    assert isinstance(validator, ProjectStructureValidator)


@pytest.mark.xfail
def test_load_validator_class_params_withparams():
    """Check that _load_validator_class_params correctly applies kwargs"""
    raise NotImplementedError


@pytest.mark.xfail
def test_check_project_structure():
    raise NotImplementedError


@pytest.mark.xfail
def test_validate_feature_api():
    raise NotImplementedError


@pytest.mark.xfail
def test_evaluate_feature_performance():
    raise NotImplementedError


@pytest.mark.xfail
def test_prune_existing_features():
    raise NotImplementedError


@pytest.mark.xfail
def test_validate():
    raise NotImplementedError



import pathlib
from unittest.mock import patch

from ballet.templating import (
    _stringify_path, render_feature_template, render_project_template,)


def test__stringify_path_pathlike():
    obj = pathlib.Path()
    output = _stringify_path(obj)
    assert isinstance(output, str)


def test__stringify_path_not_pathlike():
    obj = object()
    output = _stringify_path(obj)
    assert not isinstance(output, str)


@patch('ballet.templating.cookiecutter')
def test_render_project_template(mock_cookiecutter):
    render_project_template()

    args, _ = mock_cookiecutter.call_args

    assert len(args) == 1
    path = args[0]
    assert 'project_template' in str(path)


@patch('ballet.templating.push_branches_to_remote')
@patch('ballet.util.git.create_github_repo')
@patch('ballet.templating.Project.from_path')
@patch('ballet.templating.cookiecutter')
def test_render_project_template_create_remote(
    mock_cookiecutter, mock_from_path, mock_create_github_repo,
    mock_push_branches_to_remote,
):
    render_project_template(create_github_repo=True, github_token='token')
    mock_create_github_repo.assert_called_once()
    mock_push_branches_to_remote.assert_called_once()


@patch('ballet.templating.cookiecutter')
def test_render_feature_template(mock_cookiecutter):
    render_feature_template()

    args, _ = mock_cookiecutter.call_args

    assert len(args) == 1
    path = args[0]
    assert 'feature_template' in str(path)



from ballet.project import load_config
from ballet.util.io import load_table_from_config
from funcy import some, where


def load_data(split='train', input_dir=None):
    """Load data

    If input dir is not None, then load whatever dataset appears in
    `input_dir`. Otherwise, load the data split indicated by `split`.
    """
    if input_dir is not None:
        config = load_config()
        tables = config.get('data.tables')

        entities_table_name = config.get('data.entities_table_name')
        entities_config = some(where(tables, name=entities_table_name))
        X = load_table_from_config(input_dir, entities_config)

        targets_table_name = config.get('data.targets_table_name')
        targets_config = some(where(tables, name=targets_table_name))
        y = load_table_from_config(input_dir, targets_config)
        return X, y

    raise NotImplementedError



from ballet.validation.common import subsample_data_for_validation


def validate_feature_acceptance(accepter_class, feature, features, X_df, y_df,
                                X_df_val, y_val, subsample):
    if subsample:
        X_df, y_df, X_df_val, y_val = subsample_data_for_validation(
            X_df, y_df, X_df_val, y_val)
    accepter = accepter_class(X_df, y_df, X_df_val, y_val, features, feature)
    return accepter.judge()



from typing import Iterable

from sklearn.pipeline import FeatureUnion

from ballet.eng.base import GroupedFunctionTransformer

__all__ = (
    'SingleLagger',
    'make_multi_lagger'
)


class SingleLagger(GroupedFunctionTransformer):
    """Transformer that applies a lag operator to each group

    Args:
        lag: lag to apply
        groupby_kwargs: keyword arguments to pd.DataFrame.groupby
    """

    def __init__(self, lag: int, groupby_kwargs: dict = None):
        super().__init__(lambda x: x.shift(lag), groupby_kwargs=groupby_kwargs)


def make_multi_lagger(
    lags: Iterable[int], groupby_kwargs: dict = None
) -> FeatureUnion:
    """Return a union of transformers that apply different lags

    Args:
        lags: collection of lags to apply
        groupby_kwargs: keyword arguments to pd.DataFrame.groupby
    """
    laggers = [
        SingleLagger(lag, groupby_kwargs=groupby_kwargs)
        for lag in lags
    ]
    feature_union = FeatureUnion([
        (repr(lagger), lagger) for lagger in laggers
    ])
    return feature_union



import numpy as np
import pandas as pd
import pytest

import ballet.eng.ts
from ballet.util.testing import assert_frame_equal, assert_series_equal


def test_single_lagger():
    # simple test
    data = pd.util.testing.makeTimeSeries()

    trans = ballet.eng.ts.SingleLagger(1)
    result = trans.fit_transform(data)
    expected_result = data.shift(1)

    assert_series_equal(result, expected_result)

    data = pd.DataFrame(
        data={
            'city': ['LA', 'LA', 'LA', 'NYC', 'BOS', 'BOS', 'BOS'],
            'year': [2001, 2002, 2003, 2002, 2003, 2004, 2005],
            'width': [1, 2, 3, 4, 5, 6, 7],
        }
    ).set_index(['city', 'year']).sort_index()
    trans = ballet.eng.ts.SingleLagger(
        1, groupby_kwargs={'level': 'city'})
    result = trans.fit_transform(data)
    expected_result = pd.DataFrame(
        data={
            'city': ['LA', 'LA', 'LA', 'NYC', 'BOS', 'BOS', 'BOS'],
            'year': [2001, 2002, 2003, 2002, 2003, 2004, 2005],
            'width': [np.nan, 1, 2, np.nan, np.nan, 5, 6],
        }
    ).set_index(['city', 'year']).sort_index()

    assert_frame_equal(result, expected_result)


@pytest.mark.xfail
def test_multi_lagger(self):
    raise NotImplementedError



import random
from contextlib import contextmanager
from unittest.util import _common_shorten_repr

import funcy
import numpy as np
import numpy.testing as npt
import pandas as pd
import pandas.testing as pdt

EPSILON = 1e-4


@funcy.contextmanager
def seeded(seed):
    """Set seed, run code, then restore rng state"""
    if seed is not None:
        np_random_state = np.random.get_state()
        random_state = random.getstate()
        np.random.seed(seed)
        random.seed(seed)

    yield

    if seed is not None:
        np.random.set_state(np_random_state)
        random.setstate(random_state)


@funcy.contextmanager
def log_seed_on_error(logger, seed=None):
    """Store seed, run code, and report seed if error"""
    if seed is None:
        seed = random.randint(0, 2**32 - 1)
    try:
        with seeded(seed):
            yield
    except Exception:
        logger.exception(f'Error was thrown using seed {seed}')


@contextmanager
def _invert_assertion():
    try:
        yield
    except AssertionError:
        pass
    else:
        raise AssertionError


def assert_array_equal(first, second):
    npt.assert_array_equal(first, second, verbose=False)


def assert_array_not_equal(first, second):
    with _invert_assertion():
        npt.assert_array_equal(first, second, verbose=False)


def assert_array_almost_equal(first, second, places=6, delta=None):
    """Test that arrays first and second are almost equal

    There are two ways to control the tolerance of the approximate
    equality test, using ``places`` or ``delta``. All values are compared
    element-wise.
    ``delta``
    is provided, the allowed tolerance between the values is set to
    ``delta. Else if ``places`` is provided, values are compared up to
    ``places`` places. Else, ``places`` defaults to 6.

    Args:
        first: first object
        second: second object
        places (int, default=6): elements are compared up to this many
            places
        delta (float): the allowed tolerance between elements
    """

    # convert between both "places" and "delta" allowed to just "decimal"
    # numpy says abs(desired-actual) < 1.5 * 10**(-decimal)
    # => delta = 1.5 * 10**(-decimal)
    # => delta / 1.5 = 10**(-decimal)
    # => log10(delta/1.5) = -decimal
    # => decimal = -log10(delta) + log10(1.5)
    if delta is not None:
        places = int(-np.log10(delta) + np.log10(1.5))

    npt.assert_array_almost_equal(
        first, second, decimal=places, verbose=False)


def assert_array_not_almost_equal(
        first, second, places=6, delta=None):
    """Test that arrays first and second are not almost equal"""

    if delta is not None:
        places = int(-np.log10(delta) + np.log10(1.5))

    with _invert_assertion():
        npt.assert_array_almost_equal(
            first, second, decimal=places, verbose=False)


def assert_frame_equal(first, second, **kwargs):
    """Test that DataFrames first and second are equal"""
    _assert_pandas_equal(
        pdt.assert_frame_equal, first, second, **kwargs)


def assert_frame_not_equal(first, second, **kwargs):
    """Test that DataFrames first and second are not equal"""
    _assert_pandas_not_equal(
        pdt.assert_frame_equal, first, second, **kwargs)


def assert_series_equal(first, second, **kwargs):
    """Test that Series first and second are equal"""
    _assert_pandas_equal(
        pdt.assert_series_equal, first, second, **kwargs)


def assert_series_not_equal(first, second, **kwargs):
    """Test that Series first and second are not equal"""
    _assert_pandas_not_equal(
        pdt.assert_series_equal, first, second, **kwargs)


def assert_index_equal(first, second, **kwargs):
    """Test that Index first and second are equal"""
    _assert_pandas_equal(
        pdt.assert_index_equal, first, second, **kwargs)


def assert_index_not_equal(first, second, **kwargs):
    """Test that Index first and second are not equal"""
    _assert_pandas_not_equal(
        pdt.assert_index_equal, first, second, **kwargs)


_is_pdobj = funcy.isa(pd.core.base.PandasObject)


def assert_pandas_object_equal(first, second, **kwargs):
    """Test that arbitrary Pandas objects first and second are equal"""
    if _is_pdobj(first) and _is_pdobj(second):
        if isinstance(first, type(second)):
            if isinstance(first, pd.DataFrame):
                assert_frame_equal(first, second, **kwargs)
            elif isinstance(first, pd.Series):
                assert_series_equal(first, second, **kwargs)
            elif isinstance(first, pd.Index):
                assert_index_equal(first, second, **kwargs)
            else:
                # unreachable?
                raise AssertionError('you found a bug: unreachable code')
    else:
        msg = '{} and {} are uncomparable types'.format(
            *_common_shorten_repr(first, second))
        raise AssertionError(msg)


def assert_pandas_object_not_equal(first, second, **kwargs):
    """Test that arbitrary Pandas objects first and second are not equal"""
    if _is_pdobj(first) and _is_pdobj(second):
        if isinstance(first, type(second)):
            if isinstance(first, pd.DataFrame):
                assert_frame_not_equal(first, second, **kwargs)
            elif isinstance(first, pd.Series):
                assert_series_not_equal(first, second, **kwargs)
            elif isinstance(first, pd.Index):
                assert_index_not_equal(first, second, **kwargs)
            else:
                # unreachable?
                raise AssertionError('you found a bug: unreachable code')
    else:
        # it's great that they are uncomparable types :)
        pass


def _assert_pandas_equal(func, first, second, **kwargs):
    func(first, second, **kwargs)


def _assert_pandas_not_equal(func, first, second, **kwargs):
    try:
        func(first, second, **kwargs)
    except AssertionError:
        pass
    else:
        raise AssertionError



from abc import ABCMeta, abstractmethod
from typing import Iterable, List, Optional, Union

import numpy as np
import pandas as pd

from ballet.feature import Feature


class BaseValidator(metaclass=ABCMeta):
    """Base class for a generic validator"""

    @abstractmethod
    def validate(self) -> bool:
        """Validate something and return whether validation succeeded"""
        pass


class FeaturePerformanceEvaluator(metaclass=ABCMeta):
    """Evaluate the performance of features from an ML point-of-view

    Args:
        X_df: entities frame for fitting the features
        y_df: targets frame/series for fitting the features
        X_df_val: entities frame for evaluating the features
        y_val: target values for evaluating the features
        features: all collected features
        candidate_feature: the feature to evaluate
    """

    def __init__(self,
                 X_df: pd.DataFrame,
                 y_df: Union[pd.DataFrame, pd.Series],
                 X_df_val: pd.DataFrame,
                 y_val: np.ndarray,
                 features: Iterable[Feature],
                 candidate_feature: Feature):
        self.X_df = X_df
        self.y_df = y_df
        self.X_df_val = X_df_val
        self.y_val = y_val
        self.features = features
        self.candidate_feature = candidate_feature

    def __str__(self):
        return self.__class__.__name__


class FeatureAcceptanceMixin(metaclass=ABCMeta):

    @abstractmethod
    def judge(self) -> bool:
        """Judge whether feature should be accepted"""
        pass


class FeaturePruningMixin(metaclass=ABCMeta):

    @abstractmethod
    def prune(self) -> List[Feature]:
        """Prune existing features, returning list of features to remove"""
        pass


class FeatureAccepter(FeatureAcceptanceMixin, FeaturePerformanceEvaluator):
    """Accept/reject a feature to the project based on its performance"""
    pass


class FeaturePruner(FeaturePruningMixin, FeaturePerformanceEvaluator):
    """Prune features after acceptance based on their performance"""
    pass


class BaseCheck(metaclass=ABCMeta):

    def do_check(self, obj) -> bool:
        """Do the check and return whether an exception was *not* thrown"""
        try:
            self.check(obj)
        except Exception:
            return False
        else:
            return True

    @abstractmethod
    def check(self, obj) -> None:
        """Check something and throw an exception if the thing is bad"""
        pass

    def give_advice(self, feature) -> Optional[str]:
        """Description of how to resolve if the check fails"""
        return None



from typing import Dict, List, Optional

import funcy as fy
import numpy as np
import pandas as pd

import ballet
from ballet.transformer import get_transformer_primitives
from ballet.util import asarray2d
from ballet.validation.entropy import (
    estimate_conditional_information, estimate_mutual_information,)


def countunique(z: np.ndarray, axis=0):
    return np.apply_along_axis(
        lambda arr: len(np.unique(arr)), axis, z)


@fy.memoize(key_func=lambda feature, values, y: id(feature))
def _summarize_feature(
    feature: 'ballet.feature.Feature',
    values: Optional[Dict['ballet.feature.Feature', Optional[np.ndarray]]],
    y: Optional[np.ndarray],
) -> dict:
    """Summarize a single feature"""
    result = {
        'name': feature.name,
        'description': feature.description,
        'input':
            [feature.input]
            if isinstance(feature.input, str)
            else feature.input,
        'transformer': feature.transformer,
        'primitives': get_transformer_primitives(feature.transformer),
        'output': feature.output,
        'author': feature.author,
        'source': feature.source,
        'mutual_information': np.nan,
        'conditional_mutual_information': np.nan,
        'mean': np.nan,
        'std': np.nan,
        'variance': np.nan,
        'nunique': np.nan,
    }

    # if feature values are missing here, the values are left at nans from
    # above
    if values is not None and y is not None:
        z = values[feature]
        if z is not None:
            feature_values_list = [
                feature_values
                for other_feature, feature_values in values.items()
                if other_feature is not feature and feature_values is not None
            ]
            if feature_values_list:
                x = np.concatenate(feature_values_list, axis=1)
            else:
                x = np.empty((z.shape[0], 0))

            result['mutual_information'] = estimate_mutual_information(z, y)
            result['conditional_mutual_information'] = \
                estimate_conditional_information(z, y, x)
            result['mean'] = np.mean(np.mean(z, axis=0))  # same thing anyway
            result['std'] = np.mean(np.std(z, axis=0))
            result['variance'] = np.mean(np.var(z, axis=0))
            result['nunique'] = np.mean(countunique(z, axis=0))

    return result


def discover(
    features: List['ballet.feature.Feature'],
    X_df: Optional[pd.DataFrame],
    y_df: Optional[pd.DataFrame],
    y: Optional[np.ndarray],
    input: Optional[str] = None,
    primitive: Optional[str] = None
) -> pd.DataFrame:
    """Discover existing features

    Display information about existing features including summary statistics on
    the development dataset.  If the feature extracts multiple feature values,
    then the summary statistics (e.g. mean, std, nunique) are computed for each
    feature value and then averaged. If the development dataset cannot be
    loaded, computation of summary statistics is skipped.

    The following information is shown:
    - name: the name of the feature
    - description: the description of the feature
    - input: the variables that are used as input to the feature
    - transformer: the transformer/transformer pipeline
    - output: the output columns of the feature (not usually specified)
    - author: the GitHub username of the feature's author
    - source: the fully-qualified name of the Python module that contains the
        feature
    - mutual_information: estimated mutual information between the feature (or
        averaged over feature values) and the target on the development
        dataset split
    - conditional_mutual_information: estimated conditional mutual information
        between the feature (or averaged over feature values) and the target
        conditional on all other features on the development dataset split
    - mean: mean of the feature on the development dataset split
    - std: standard deviation of the feature (or averaged over feature values)
        on the development dataset split
    - var: variance of the feature (or averaged over feature values) on the
        development dataset split
    - nunique: number of unique values of the feature (or averaged over
        feature values) on the development dataset split

    The following query operators are supported:
    - input (str): filter to only features that have ``input`` in their input/
        list of inputs
    - primitive (str): filter to only features that use primitive
        ``primitive`` (i.e. a class with name ``primitive``) in the
        transformer/transformer pipeline

    For other queries, you should just use normal DataFrame indexing::

       >>> features_df[features_df['author'] == 'jane']
       >>> features_df[features_df['name'].str.contains('married')]
       >>> features_df[features_df['mutual_information'] > 0.05]
       >>> features_df[features_df['input'].apply(
               lambda input: 'A' in input and 'B' in input)]

    Returns:
        data frame with features on the row index and columns as described
        above
    """
    records = []

    if X_df is not None and y_df is not None and y is not None:

        @fy.ignore(Exception)
        def get_feature_values(feature):
            return asarray2d(
                feature
                .as_feature_engineering_pipeline()
                .fit_transform(X_df, y_df))

        values = {
            feature: get_feature_values(feature)
            for feature in features
        }
        y = asarray2d(y)
        summarize = fy.rpartial(_summarize_feature, values, y)

    else:
        summarize = fy.rpartial(_summarize_feature, None, None)

    for feature in features:
        if input and input not in feature.input and input != feature.input:
            continue
        if (
            primitive
            and primitive not in get_transformer_primitives(
                feature.transformer)
        ):
            continue
        summary = summarize(feature)
        records.append(summary)

    return pd.DataFrame.from_records(records)



import importlib
from textwrap import dedent
from types import ModuleType
from unittest.mock import create_autospec, patch

import git
import pytest

from ballet.project import Project
from ballet.util.git import CustomDiffer, Differ
from ballet.validation.common import ChangeCollector, NewFeatureInfo
from ballet.validation.feature_api.validator import FeatureApiValidator
from ballet.validation.project_structure.validator import (
    ProjectStructureValidator,)

from ..util import make_mock_commit, make_mock_commits


@pytest.fixture
def invalid_feature_code():
    return dedent(
        '''
        from sklearn.base import BaseEstimator, TransformerMixin
        class RaisingTransformer(BaseEstimator, TransformerMixin):
            def fit(self, X, y=None, **fit_kwargs):
                raise RuntimeError
            def transform(self, X, **transform_kwargs):
                raise RuntimeError
        input = 'size'
        transformer = RaisingTransformer()
        '''
    ).strip()


@pytest.fixture
def import_error_code():
    return dedent(
        '''
        edf foo():
            pass
        '''
    ).strip()


def code_to_module(code: str, modname='modname') -> ModuleType:
    # see https://stackoverflow.com/a/53080237
    spec = importlib.util.spec_from_loader(modname, loader=None)
    module = importlib.util.module_from_spec(spec)
    exec(code, module.__dict__)
    return module


def test_change_collector_init():
    project = create_autospec(Project)
    differ = Differ()
    change_collector = ChangeCollector(project, differ=differ)
    assert change_collector is not None


@patch('ballet.validation.common.can_use_travis_differ', return_value=True)
@patch('ballet.validation.common.TravisPullRequestBuildDiffer')
def test_change_collector_detect_differ_travis(mock_travis_differ, _):
    """Check ChangeCollector._detect_differ"""
    differ_instance = mock_travis_differ.return_value
    project = create_autospec(Project)
    change_collector = ChangeCollector(project)
    assert change_collector.differ is differ_instance


def test_change_collector_collect_file_diffs_custom_differ(mock_repo):
    repo = mock_repo

    n = 10
    filename = 'file{i}.py'

    commits = make_mock_commits(repo, n=n, filename=filename)

    project = None
    differ = CustomDiffer(endpoints=(commits[0], commits[-1]))
    change_collector = ChangeCollector(project, differ=differ)
    file_diffs = change_collector._collect_file_diffs()

    # checks on file_diffs
    assert len(file_diffs) == n - 1

    for diff in file_diffs:
        assert diff.change_type == 'A'
        assert diff.b_path.startswith('file')
        assert diff.b_path.endswith('.py')


@pytest.mark.xfail
def test_change_collector_categorize_file_diffs():
    raise NotImplementedError


@pytest.mark.xfail
def test_change_collector_collect_features():
    raise NotImplementedError


def test_change_collector_collect_changes(quickstart):
    repo = quickstart.repo
    contrib_path = quickstart.project.config.get('contrib.module_path')

    path_content = [
        ('something.txt', None),  # invalid
        ('invalid.py', None),  # invalid
        (f'{contrib_path}/foo/bar/baz.py', None),  # invalid

        # candidate_feature, and also new_feature_info
        (f'{contrib_path}/user_foo/feature_bar.py', None),

        (f'{contrib_path}/user_foo/__init__.py', None),  # valid_init
    ]

    old_head = repo.head.commit

    for path, content in path_content:
        make_mock_commit(repo, path=path, content=content)

    new_head = repo.head.commit

    differ = CustomDiffer(endpoints=(old_head, new_head))
    change_collector = ChangeCollector(quickstart.project, differ=differ)
    changes = change_collector.collect_changes()

    assert len(changes.file_diffs) == 5
    assert len(changes.candidate_feature_diffs) == 1
    assert len(changes.valid_init_diffs) == 1
    assert len(changes.inadmissible_diffs) == 3
    assert len(changes.new_feature_info) == 1

    actual_inadmissible = [
        diff.b_path
        for diff in changes.inadmissible_diffs
    ]
    expected_inadmissible = [
        'something.txt', 'invalid.py', f'{contrib_path}/foo/bar/baz.py'
    ]
    assert set(actual_inadmissible) == set(expected_inadmissible)


@pytest.mark.parametrize(
    'inadmissible_diffs,expected',
    [
        ([create_autospec(git.Diff)], False),
        ([], True),
    ]
)
@patch('ballet.validation.project_structure.validator.ChangeCollector')
def test_project_structure_validator(
    mock_change_collector, inadmissible_diffs, expected,
):
    mock_change_collector \
        .return_value \
        .collect_changes \
        .return_value \
        .inadmissible_diffs = inadmissible_diffs

    project = None
    validator = ProjectStructureValidator(project)
    result = validator.validate()
    assert result == expected


@patch('ballet.validation.feature_api.validator.ChangeCollector')
def test_feature_api_validator_validation_failure_no_features_found(
    mock_change_collector,
    sample_data,
):
    """
    If the change collector does not return any new features, validation fails.
    """
    mock_change_collector \
        .return_value \
        .collect_changes \
        .return_value \
        .new_feature_info = []

    project = create_autospec(Project)
    project.api.load_data.return_value = sample_data.X, sample_data.y
    validator = FeatureApiValidator(project)
    result = validator.validate()
    assert not result


@patch(
    'ballet.validation.feature_api.validator.validate_feature_api',
    return_value=True
)
@patch('ballet.validation.feature_api.validator.ChangeCollector')
def test_feature_api_validator_validation_failure_invalid_feature(
    mock_change_collector, mock_validate_feature_api,
    sample_data, invalid_feature_code,
):
    """
    If change collector returns a feature, but validate_feature_api fails,
    then validation fails.
    """
    mock_change_collector \
        .return_value \
        .collect_changes \
        .return_value \
        .new_feature_info = [
            NewFeatureInfo(
                lambda: code_to_module(invalid_feature_code),
                'modname',
                'modpath',
            ),
        ]

    project = create_autospec(Project)
    project.api.load_data.return_value = sample_data.X, sample_data.y
    validator = FeatureApiValidator(project)
    result = validator.validate()
    assert not result


@patch('ballet.validation.feature_api.validator.ChangeCollector')
def test_feature_api_validator_validation_failure_import_error(
    mock_change_collector,
    sample_data, import_error_code,
):
    """
    If change collector returns a feature, but importer() fails to return a
    module, then validation fails.
    """
    mock_change_collector \
        .return_value \
        .collect_changes \
        .return_value \
        .new_feature_info = [
            NewFeatureInfo(
                lambda: code_to_module(import_error_code),
                'modname',
                'modpath',
            ),
        ]

    project = create_autospec(Project)
    project.api.load_data.return_value = sample_data.X, sample_data.y
    validator = FeatureApiValidator(project)
    result = validator.validate()
    assert not result



import logging

from ballet.util.io import save_features, save_targets
import click

from {{ cookiecutter.package_slug }}.api import api

logger = logging.getLogger(__name__)


@click.group()
def cli():
    import ballet.util.log
    ballet.util.log.enable(logger=logger, level='INFO', echo=False)
    ballet.util.log.enable(logger=ballet.util.log.logger, level='INFO',
                           echo=False)


@cli.command()
@click.argument('input_dir', type=click.Path(
    exists=True, file_okay=False, readable=True))
@click.argument('output_dir', type=click.Path(file_okay=False))
def engineer_features(input_dir, output_dir):
    """Engineer features"""
    X_df, y_df = api.load_data(input_dir=input_dir)
    result = api.engineer_features()
    pipeline, encoder = result.pipeline, result.encoder

    X_ft = pipeline.transform(X_df, y=y_df)
    y_ft = encoder.transform(y_df)

    save_features(X_ft, output_dir)
    save_targets(y_ft, output_dir)


if __name__ == '__main__':
    cli(prog_name='{{ cookiecutter.package_slug }}')



import pathlib
import tempfile
from textwrap import dedent

from funcy import contextmanager

from ballet.contrib import _collect_contrib_features
from ballet.util.mod import import_module_at_path


def create_contrib_modules_at_dir(dirname, modcontent, n=1):
    '''Create the a "contrib" directory structure at dirname

    Creates up to mod{n}:

        dirname/
        ├── __init__.py
        ├── mod1/
        |   ├── __init__.py
        |   └── foo1.py
        ├── mod2/
        |   ├── __init__.py
        |   └── foo2.py
        etc.
    '''

    root = pathlib.Path(dirname)
    root.joinpath('__init__.py').touch()
    for i in range(n):
        root.joinpath(f'mod{i}').mkdir()
        root.joinpath(f'mod{i}', '__init__.py').touch()

        try:
            modcontent_i = modcontent.format(i=i)
        except KeyError:
            # shouldn't happen
            modcontent_i = modcontent
        modpath_i = str(root.joinpath(f'mod{i}', f'foo{i}.py'))
        with open(modpath_i, 'w') as f:
            f.write(modcontent_i)


@contextmanager
def mock_contrib_module(modname, content, n):
    with tempfile.TemporaryDirectory() as tmpdir:
        modpath = pathlib.Path(tmpdir).joinpath(modname)
        modpath.mkdir()
        create_contrib_modules_at_dir(modpath, content, n=n)
        mod = import_module_at_path(modname, modpath)
        features = _collect_contrib_features(mod)
        yield mod, features


def test_collect_contrib_features_stdlib():
    # give a nonsense *module*, shouldn't import anything. this is a bad
    # test because it relies on module not defining certain names
    import math
    features = _collect_contrib_features(math)

    # features should be an empty list
    assert len(features) == 0


def test_collect_contrib_features_thirdparty():
    # give a nonsense *package*, shouldn't import anything. this is a bad
    # test because it relies on module not defining certain names
    import funcy
    features = _collect_contrib_features(funcy)
    assert len(features) == 0


def test_collect_contrib_features_generated():
    n = 4
    content = dedent(
        '''
        from ballet import Feature
        from sklearn.preprocessing import StandardScaler
        input = 'col{i}'
        transformer = StandardScaler()
        feature = Feature(input, transformer)
        '''
    ).strip()
    modname = 'contrib_features_generated'
    with mock_contrib_module(modname, content, n) as (mod, features):
        assert len(features) == n



__all__ = []
try:
    from category_encoders import BackwardDifferenceEncoder
    __all__.append('BackwardDifferenceEncoder')

    from category_encoders import BaseNEncoder
    __all__.append('BaseNEncoder')

    from category_encoders import BinaryEncoder
    __all__.append('BinaryEncoder')

    from category_encoders import CatBoostEncoder
    __all__.append('CatBoostEncoder')

    from category_encoders import CountEncoder
    __all__.append('CountEncoder')

    from category_encoders import GLMMEncoder
    __all__.append('GLMMEncoder')

    from category_encoders import HashingEncoder
    __all__.append('HashingEncoder')

    from category_encoders import HelmertEncoder
    __all__.append('HelmertEncoder')

    from category_encoders import JamesSteinEncoder
    __all__.append('JamesSteinEncoder')

    from category_encoders import LeaveOneOutEncoder
    __all__.append('LeaveOneOutEncoder')

    from category_encoders import MEstimateEncoder
    __all__.append('MEstimateEncoder')

    from category_encoders import OneHotEncoder
    __all__.append('OneHotEncoder')

    from category_encoders import OrdinalEncoder
    __all__.append('OrdinalEncoder')

    from category_encoders import PolynomialEncoder
    __all__.append('PolynomialEncoder')

    from category_encoders import SumEncoder
    __all__.append('SumEncoder')

    from category_encoders import TargetEncoder
    __all__.append('TargetEncoder')

    from category_encoders import WOEEncoder
    __all__.append('WOEEncoder')
except ImportError:
    pass



from ballet.eng.misc import IdentityTransformer


def get_target_encoder():
    """Get encoder for the prediction target

    Returns:
        transformer-like
    """
    return IdentityTransformer()



import warnings
from contextlib import redirect_stderr, redirect_stdout, suppress
from copy import deepcopy
from os import devnull
from typing import Collection, Optional, Sequence, Tuple, TypeVar

import cookiecutter.utils
import numpy as np
import pandas as pd
import sklearn.datasets
from funcy import complement, decorator, lfilter
from funcy.decorators import Call

from ballet.exc import BalletWarning
from ballet.util.log import logger

RANDOM_STATE = 1754


def asarray2d(a: np.ndarray) -> np.ndarray:
    """Cast to 2d array"""
    arr = np.asarray(a)
    if arr.ndim == 1:
        arr = arr.reshape(-1, 1)
    return arr


def get_arr_desc(arr: np.ndarray) -> str:
    """Get array description, in the form '<array type> <array shape>'"""
    type_ = type(arr).__name__  # see also __qualname__
    shape = getattr(arr, 'shape', '<no shape>')
    return f'{type_} {shape}'


def indent(text: str, n=4) -> str:
    """Indent each line of text by n spaces"""
    _indent = ' ' * n
    return '\n'.join(_indent + line for line in text.split('\n'))


def make_plural_suffix(obj: Collection, suffix='s') -> str:
    if len(obj) != 1:
        return suffix
    else:
        return ''


def has_nans(obj) -> bool:
    """Check if obj has any NaNs

    Compatible with different behavior of np.isnan, which sometimes applies
    over all axes (py35+) and sometimes does not (py34).
    """
    nans = np.isnan(obj)
    while np.ndim(nans):
        nans = np.any(nans)
    return bool(nans)


@decorator
def dfilter(call: Call, pred):
    """Decorate a callable with a filter that accepts a predicate

    Example::

        >>> @dfilter(lambda x: x >= 0)
        ... def numbers():
        ...     return [-1, 2, 0, -2]
        [2, 0]
    """
    return lfilter(pred, call())


def load_sklearn_df(name: str) -> Tuple[pd.DataFrame, pd.DataFrame]:
    method_name = f'load_{name}'
    method = getattr(sklearn.datasets, method_name)
    data = method()
    X_df = pd.DataFrame(data=data.data, columns=data.feature_names)
    y_df = pd.Series(data.target, name='target')
    return X_df, y_df


@decorator
def quiet(call: Call):
    with open(devnull, 'w') as fnull:
        with redirect_stderr(fnull), redirect_stdout(fnull):
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                return call()


class DeepcopyMixin:

    def __deepcopy__(self, memo: dict):
        cls = self.__class__
        result = cls.__new__(cls)
        memo[id(self)] = result
        for k, v in self.__dict__.items():
            setattr(result, k, deepcopy(v, memo))
        return result


_T = TypeVar('_T')


def one_or_raise(seq: Sequence[_T]) -> _T:
    n = len(seq)
    if n == 1:
        return seq[0]
    else:
        raise ValueError(f'Expected exactly 1 element, but got {n}')


def warn(msg: str):
    """Issue a warning message of category BalletWarning"""
    warnings.warn(msg, category=BalletWarning)


@decorator
def raiseifnone(call: Call):
    """Decorate a function to raise a ValueError if result is None"""
    result = call()
    if result is None:
        raise ValueError
    else:
        return result


def falsy(o) -> bool:
    """Check whether o is falsy

    In this case, a falsy value is one of the following:
    1. the singleton value `False`
    2. the string 'false' (ignoring case)
    3. the empty string
    """
    if isinstance(o, bool):
        return not o
    return isinstance(o, str) and (o.lower() == 'false' or o == '')


truthy = complement(falsy)
"""Check whether o is truthy

In this case, a truthy value is any value that is not falsy.
"""


@decorator
def nonnegative(call: Call, name: Optional[str] = None):
    """Warn if the function's return value is negative and set it to 0"""
    result = call()
    with suppress(TypeError):
        if result < 0:
            result = 0.0
            # Format a nice log message
            if name is None:
                try:
                    pieces = call._func.__name__.split('_')[1:]
                    name = ''.join(map(str.capitalize, pieces))
                except RuntimeError:
                    name = 'Result'
            logger.warning(f'{name} should be non-negative.')
    return result


# re-export cookiecutter work_in
work_in = cookiecutter.utils.work_in



from .category_encoders import *
from .feature_engine import *
from .featuretools import *
from .skits import *
from .sklearn import *
from .tsfresh import *

# needed for sphinx
from .category_encoders import __all__ as _category_encoders_all
from .feature_engine import __all__ as _feature_engine_all
from .featuretools import __all__ as _featuretools_all
from .skits import __all__ as _skits_all
from .sklearn import __all__ as _sklearn_all
from .tsfresh import __all__ as _tsfresh_all
__all__ = (*_category_encoders_all, *_feature_engine_all,
           *_featuretools_all, *_skits_all, *_sklearn_all, *_tsfresh_all)



import random

from ballet.util.log import logger
from ballet.util.testing import seeded
from ballet.validation.base import FeaturePruner, FeaturePruningMixin
from ballet.validation.common import RandomFeaturePerformanceEvaluator
from ballet.validation.entropy import estimate_conditional_information
from ballet.validation.gfssf import (
    GFSSFPerformanceEvaluator, _compute_lmbdas, _compute_threshold,
    _concat_datasets,)


class NoOpPruner(FeaturePruner):
    def prune(self):
        logger.info(f'Pruning features using {self}')
        return []


class RandomPruner(FeaturePruningMixin, RandomFeaturePerformanceEvaluator):

    def prune(self):
        """With probability p, select a random feature to prune"""
        logger.info(f'Pruning features using {self}')
        with seeded(self.seed):
            if random.uniform(0, 1) < self.p:
                return [random.choice(self.features)]


CMI_MESSAGE = 'Calculating CMI of feature and target cond. on accpt features'


class GFSSFPruner(FeaturePruningMixin, GFSSFPerformanceEvaluator):

    def prune(self):
        """Prune using GFSSF

        Uses lines 12-13 of agGFSSF
        """

        logger.info(f'Pruning features using {self}')

        feature_df_map = self._get_feature_df_map()
        lmbda_1, lmbda_2 = _compute_lmbdas(
            self.lmbda_1, self.lmbda_2, feature_df_map)

        logger.info(f'Recomputed lambda_1={lmbda_1}, lambda_2={lmbda_2}')

        redundant_features = []
        for candidate_feature in self.features:
            candidate_src = candidate_feature.source
            logger.debug(
                f'Trying to prune feature with source {candidate_src}')
            candidate_df = feature_df_map[candidate_feature]
            _, n_candidate_cols = candidate_df.shape
            z = _concat_datasets(feature_df_map, omit=[candidate_feature])
            logger.debug(CMI_MESSAGE)
            cmi = estimate_conditional_information(candidate_df, self.y_val, z)

            logger.debug(f'Conditional Mutual Information Score: {cmi}')
            statistic = cmi
            threshold = _compute_threshold(lmbda_1, lmbda_2, n_candidate_cols)
            logger.debug(f'Calculated Threshold: {threshold}')
            if statistic >= threshold:
                logger.debug(f'Passed, keeping feature {candidate_src}')
            else:
                # ballet.validation.main._prune_existing_features will log
                # this at level INFO
                logger.debug(
                    f'Failed, found redundant feature: {candidate_src}')
                del feature_df_map[candidate_feature]
                redundant_features.append(candidate_feature)
        return redundant_features



import inspect
import pathlib
import sys
from functools import partial
from importlib import import_module
from types import ModuleType
from typing import Any, Callable, List, Optional, Tuple

import git
from dynaconf import Dynaconf
from funcy import cache, cached_property, fallback, re_find
from pandas import DataFrame

import ballet.contrib
from ballet.compat import is_mount
from ballet.eng import BaseTransformer
from ballet.exc import ConfigurationError
from ballet.feature import Feature
from ballet.pipeline import (
    EngineerFeaturesResult, FeatureEngineeringPipeline,
    make_engineer_features,)
from ballet.util import raiseifnone
from ballet.util.ci import get_travis_branch
from ballet.util.git import get_branch
from ballet.util.mod import import_module_at_path
from ballet.util.typing import Pathy

DEFAULT_CONFIG_NAME = 'ballet.yml'
DYNACONF_OPTIONS = {
    'envvar_prefix': 'BALLET',
    'settings_file': DEFAULT_CONFIG_NAME,
    'yaml_loader': 'safe_load',
}


def load_config(path: Optional[Pathy] = None, ascend: bool = True) -> Dynaconf:
    """User-facing function to load config from project code

    The default behavior when no arguments are provided is to detect the
    calling code using introspection and load a config object by ascending
    the directory of the calling code. If this does not succeed, you should
    just pass `path` directly.
    """
    if path is None:
        # "The first entry in the returned list represents the caller; the
        # last entry represents the outermost call on the stack."
        frame = inspect.stack()[1]
        try:
            path = frame.filename
        finally:
            del frame

    if path in ('<stdin>', '<string>'):
        path = pathlib.Path.cwd()

    path = pathlib.Path(path)
    while path.exists() and not is_mount(path):
        try:
            return load_config_in_dir(path)
        except ConfigurationError:
            if ascend:
                path = path.parent

    raise ConfigurationError


def load_config_at_path(path: Pathy) -> Dynaconf:
    """Load config at exact path

    Args:
        path: path to config file

    Returns:
        dict: config dict
    """
    path = pathlib.Path(path)
    if path.exists() and path.is_file():
        options = DYNACONF_OPTIONS.copy()
        options.update({
            'root_path': str(path.parent),
            'settings_file': str(path.name),
        })
        return Dynaconf(**options)
    else:
        raise ConfigurationError(
            f'Couldn\'t find ballet.yml config file at {path!s}')


def load_config_in_dir(path: Pathy) -> Dynaconf:
    """Load config in containing directory

    Args:
        path: path to containing directory of config file

    Returns:
        config dict
    """
    path = pathlib.Path(path)
    return load_config_at_path(path.joinpath(DEFAULT_CONFIG_NAME))


def relative_to_contrib(
    diff: git.diff.Diff, project: 'Project'
) -> pathlib.Path:
    """Compute relative path of changed file to contrib dir

    Args:
        diff: file diff
        project: project
    """
    path = pathlib.Path(diff.b_path)
    contrib_path = project.config.get('contrib.module_path')
    return path.relative_to(contrib_path)


def make_feature_path(
    contrib_dir: Pathy, username: str, featurename: str
) -> pathlib.Path:
    contrib_dir = pathlib.Path(contrib_dir)
    return contrib_dir.joinpath(
        f'user_{username}', f'feature_{featurename}.py')


def detect_github_username(project: 'Project') -> str:
    """Detect github username

    Looks in the following order:
    1. github.user git config variable
    2. git remote origin
    3. $USER
    4. 'username'
    """
    @raiseifnone
    def get_config_variable():
        return project.repo.config_reader().get_value(
            'github', 'user', default=None)

    @raiseifnone
    def get_remote():
        url = list(project.repo.remote('origin').urls)[0]
        # protocol:user/repo, i.e. 'git@github.com:ballet/ballet'
        return re_find(r'.+:(.+)/.+', url)

    @raiseifnone
    def get_user_env():
        return sys.environ.get('USER')

    def get_default():
        return 'username'

    return fallback(
        get_config_variable, get_remote, get_user_env, (get_default, ())
    )


class Project:
    """Encapsulate information on a ballet project

    This is a utility class mostly useful for easy access to the project's
    information from within the ballet.validation package.

    Args:
        package (ModuleType): python package representing imported ballet
            project
    """

    def __init__(self, package: ModuleType):
        self.package = package

    @cached_property
    def config(self) -> Dynaconf:
        return load_config_in_dir(self.path)

    @classmethod
    def from_path(cls, path: Pathy, ascend: bool = False):
        """Create a Project instance from an fs path to the containing dir

        Args:
            path: path to directory that contains the project
            ascend: if the config file is not found in the given directory,
                then search in parent directories, stopping at a file system
                boundary
        """
        path = pathlib.Path(path)
        try:
            config = load_config_in_dir(path)
            package_slug = config.get('project.package_slug')
            package = import_module_at_path(package_slug,
                                            path.joinpath('src', package_slug))
            return cls(package)
        except ConfigurationError:
            if ascend:
                parent = path.parent
                if parent.exists() and not is_mount(parent):
                    return cls.from_path(parent, ascend=ascend)
            raise

    @classmethod
    def from_cwd(cls):
        """Create a Project instance by searching up from cwd

        Recursively searches for the ballet configuration file at the
        current working directory and parent directories, stopping when it
        reaches a file system boundary.

        Raises:
            ConfigurationError: couldn't find the configuration file
        """
        cwd = pathlib.Path.cwd()
        return cls.from_path(cwd, ascend=True)

    def resolve(
        self, modname: str, attr: str = None
    ) -> Any:
        """Import module or attribute from project

        Args:
            modname: dotted module name relative to top-level with leading
                dot omited; if trying to import the top-level package,
                use '' (can also just access self.package)
            attr: attribute to get from the imported module

        Example:

            >>> project.resolve('', '__version__')
            # return __version__ attribute from top-level package
            >>> project.resolve('api')
            # return myproject.api module
            >>> project.resolve('api', attr='api')
            # return api object from myproject.api module
            >>> project.resolve('foo.bar')
            # return myproject.foo.bar module
        """

        if modname:
            module = import_module('.' + modname,
                                   package=self.package.__name__)
        else:
            module = self.package

        if attr is not None:
            return getattr(module, attr)
        else:
            return module

    @property
    def branch(self) -> Optional[str]:
        """Return current git branch according to git tree or CI environment"""
        result = get_branch(repo=self.repo)
        if result is None:
            result = get_travis_branch()
        return result

    @property
    def on_master(self) -> bool:
        return self.branch == 'master'

    @cached_property
    def path(self) -> pathlib.Path:
        """Return the project path (aka project root)

        If ``package.__file__`` is ``/foo/src/foo/__init__.py``,
        then project.path
        should be ``/foo``.
        """
        return pathlib.Path(self.package.__file__).resolve().parents[2]

    @cached_property
    def repo(self) -> git.Repo:
        """Return a git.Repo object corresponding to this project"""
        return git.Repo(self.path, search_parent_directories=True)

    @property
    def api(self) -> 'FeatureEngineeringProject':
        return self.resolve('api', 'api')

    @property
    def version(self) -> str:
        """Some version identifier for the current project

        Implementation is to return the abbreviated SHA1 of git HEAD.
        """
        return self.repo.head.commit.hexsha[:7]


class FeatureEngineeringProject:

    CACHE_TIMEOUT = 10 * 60

    def __init__(
        self,
        *,
        package: ModuleType,
        encoder: BaseTransformer,
        load_data: Callable[..., Tuple[DataFrame, DataFrame]],
        extra_features: List[Feature] = None,
        engineer_features: Callable[..., EngineerFeaturesResult] = None,
    ):
        self._package = package
        self.encoder = encoder
        self._load_data = cache(self.CACHE_TIMEOUT)(load_data)
        self._extra_features = extra_features or []
        self._engineer_features = engineer_features

        # use fqn to avoid circular import
        self.collect = partial(ballet.contrib.collect_contrib_features,
                               self.project)

    @cached_property
    def project(self) -> Project:
        """Get the Project object representing this project."""
        return Project(self._package)

    @property
    def features(self) -> List[Feature]:
        """Get all features from the project

        Both collects all contrib features from the project and allows extra
        features to be provided by the API author.
        """
        return self.collect() + self._extra_features

    @property
    def pipeline(self) -> FeatureEngineeringPipeline:
        """Get the feature engineering pipeline from the existing features"""
        return FeatureEngineeringPipeline(self.features)

    def load_data(self, *args, cache=True, **kwargs):
        """Call the project's load_data function, caching dataset

        Dataset is cached for `FeatureEngineeringProject.CACHE_TIMEOUT`
        seconds. To invalidate cache and cause data to be re-loaded from
        wherever it comes from, pass `cache=False`.
        """
        if not cache:
            self._load_data.invalidate_all()
        return self._load_data(*args, **kwargs)

    def engineer_features(self, *args, **kwargs) -> EngineerFeaturesResult:
        """Engineer features"""
        if self._engineer_features is None:
            self._engineer_features = make_engineer_features(
                self.pipeline, self.encoder, self.load_data)

        return self._engineer_features(*args, **kwargs)



import io
from copy import deepcopy

import dill as pickle
import numpy as np
from funcy import all, all_fn, isa, iterable
from sklearn.model_selection import train_test_split

from ballet.feature import Feature
from ballet.util import RANDOM_STATE
from ballet.validation.base import BaseCheck


def _get_one_row(*args):
    return tuple(
        obj.iloc[0:1]
        for obj in args
    )


class FeatureApiCheck(BaseCheck):
    """Base class for implementing new Feature API checks

    Args:
        X_df (array-like): X dataframe
        y_df (array-like): y dataframe
    """

    def __init__(self, X_df, y_df):
        self.X = X_df
        self.y = y_df


class IsFeatureCheck(FeatureApiCheck):

    def check(self, feature):
        """Check that the object is an instance of ballet.Feature"""
        assert isinstance(feature, Feature)

    def give_advice(self, feature):
        return f'The object needs to be an instance of ballet.Feature, whereas it is actually of type {type(feature).__name__}'  # noqa


class HasCorrectInputTypeCheck(FeatureApiCheck):

    def check(self, feature):
        """Check that the feature's `input` is a str or Iterable[str]"""
        input = feature.input
        is_str = isa(str)
        is_nested_str = all_fn(
            iterable, lambda x: all(is_str, x))
        assert is_str(input) or is_nested_str(input)

    def give_advice(self, feature):
        return f'The feature\'s input needs to be a string or list of strings, whereas it is actually of type {type(feature.input).__name__}'  # noqa


class HasTransformerInterfaceCheck(FeatureApiCheck):

    def check(self, feature):
        """Check that the feature has a fit/transform/fit_tranform interface"""
        assert hasattr(feature.transformer, 'fit')
        assert hasattr(feature.transformer, 'transform')
        assert hasattr(feature.transformer, 'fit_transform')

    def give_advice(self, feature):
        missing = ', '.join(
            attr
            for attr in ('fit', 'transform', 'fit_transform')
            if not hasattr(feature.transformer, attr)
        )
        return f'The feature\'s transformer must have the transformer interface, but these methods are missing and must be implemented: {missing}'  # noqa


class CanMakeMapperCheck(FeatureApiCheck):

    def check(self, feature):
        """Check that the feature can be converted to a FEP"""
        feature.as_feature_engineering_pipeline()

    def give_advice(self, feature):
        return 'The following method call fails and needs to be fixed: feature.as_feature_engineering_pipeline()'  # noqa


class CanFitCheck(FeatureApiCheck):

    def check(self, feature):
        """Check that fit can be called on reference data"""
        mapper = feature.as_feature_engineering_pipeline()
        mapper.fit(self.X, y=self.y)

    def give_advice(self, feature):
        return 'The feature fails when calling fit on sample data'


class CanFitOneRowCheck(FeatureApiCheck):

    def check(self, feature):
        """Check that fit can be called on one row of reference data"""
        mapper = feature.as_feature_engineering_pipeline()
        x, y = _get_one_row(self.X, self.y)
        mapper.fit(x, y=y)

    def give_advice(self, feature):
        return 'The feature fails when calling fit on a single row of sample data (i.e. 1xn array)'  # noqa


class CanTransformCheck(FeatureApiCheck):

    def check(self, feature):
        """Check that transform can be called on reference data"""
        mapper = feature.as_feature_engineering_pipeline()
        mapper.fit(self.X, y=self.y)
        mapper.transform(self.X)

    def give_advice(self, feature):
        return 'The feature fails when calling transform on sample data'


class CanTransformNewRowsCheck(FeatureApiCheck):

    def check(self, feature):
        """Check that transform can be called on new, unseen rows"""
        mapper = feature.as_feature_engineering_pipeline()
        X1, X2, y1, _ = train_test_split(
            self.X, self.y, test_size=0.1, random_state=RANDOM_STATE,
            shuffle=True)
        mapper.fit(X1, y=y1)
        mapper.transform(X2)

    def give_advice(self, feature):
        return 'The feature fails when calling transform on different data than it was trained on; make sure the transform method works on any number of new rows'  # noqa


class CanTransformOneRowCheck(FeatureApiCheck):

    def check(self, feature):
        """Check that transform can be called on one row of reference data"""
        mapper = feature.as_feature_engineering_pipeline()
        mapper.fit(self.X, y=self.y)
        x, = _get_one_row(self.X)
        mapper.transform(x)

    def give_advice(self, feature):
        return 'The feature fails when calling transform on a single row of sample data (i.e. 1xn array)'  # noqa


class CanFitTransformCheck(FeatureApiCheck):

    def check(self, feature):
        """Check that fit_transform can be called on reference data"""
        mapper = feature.as_feature_engineering_pipeline()
        mapper.fit_transform(self.X, y=self.y)

    def give_advice(self, feature):
        return 'The feature fails when calling fit_transform on sample data'


class HasCorrectOutputDimensionsCheck(FeatureApiCheck):

    def check(self, feature):
        """Check that the dimensions of the transformed data are correct

        For input X, an n x p array, a n x q array should be produced,
        where q is the number of feature values produced by the feature.
        """
        mapper = feature.as_feature_engineering_pipeline()
        X = mapper.fit_transform(self.X, y=self.y)
        assert self.X.shape[0] == X.shape[0]

    def give_advice(self, feature):
        mapper = feature.as_feature_engineering_pipeline()
        X = mapper.fit_transform(self.X, y=self.y)
        n = self.X.shape[0]
        m = X.shape[0]

        return f'The feature does not produce the correct output dimensions, for example when it is fit and transformed on {n} rows of data, it produces {m} rows of feature values.'  # noqa


class CanDeepcopyCheck(FeatureApiCheck):

    def check(self, feature):
        """Check that the feature can be deepcopied

        This is needed for execution of the overall transformation pipeline
        """
        deepcopy(feature)

    def give_advice(self, feature):
        return 'Calling copy.deepcopy(feature) fails, make sure every component of the feature can be deepcopied'  # noqa


class CanPickleCheck(FeatureApiCheck):

    def check(self, feature):
        """Check that the feature can be pickled

        This is needed for saving the pipeline to disk
        """
        try:
            buf = io.BytesIO()
            pickle.dump(feature, buf, protocol=pickle.HIGHEST_PROTOCOL)
            buf.seek(0)
            new_feature = pickle.load(buf)
            assert new_feature is not None
            assert isinstance(new_feature, Feature)
        finally:
            buf.close()

    def give_advice(self, feature):
        return 'Calling pickle.dump(feature, buf) fails, make sure the feature can be pickled'  # noqa


class NoMissingValuesCheck(FeatureApiCheck):

    def check(self, feature):
        """Check that the output of the transformer has no missing values"""
        mapper = feature.as_feature_engineering_pipeline()
        X = mapper.fit_transform(self.X, y=self.y)
        assert not np.any(np.isnan(X))

    def give_advice(self, feature):
        return 'When transforming sample data, the feature produces NaN values. If you reasonably expect these missing values, make sure you clean missing values as an additional step in your transformer list. For example: NullFiller(replacement=replacement)'  # noqa


class NoInfiniteValuesCheck(FeatureApiCheck):

    def check(self, feature):
        """Check that the output of the transformer has no non-finite values"""
        mapper = feature.as_feature_engineering_pipeline()
        X = mapper.fit_transform(self.X, y=self.y)
        assert not np.any(np.isinf(X))

    def give_advice(self, feature):
        return 'When transforming sample data, the feature produces infinite values. You can detect these with np.isinf. If you reasonably expect these infinite values, make sure you clean infinite values as an additional step in your transformer list. For example: NullFiller(np.isinf, replacement) '  # noqa



import pathlib
import types

import pytest

import ballet
from ballet.util.mod import (
    import_module_at_path, modname_to_relpath, relpath_to_modname,)


@pytest.mark.xfail
def test_import_module_from_modname():
    import ballet.util.mod.import_module_from_modname  # noqa F401
    raise NotImplementedError


@pytest.mark.xfail
def test_import_module_from_relpath():
    import ballet.util.mod.import_module_from_relpath  # noqa F401
    raise NotImplementedError


def test_import_module_at_path_module(tmp_path):
    path = tmp_path.joinpath('foo', 'bar.py')
    path.parent.mkdir(parents=True)
    init = path.parent.joinpath('__init__.py')
    init.touch()
    x = 1
    with path.open('w') as f:
        f.write(f'x={x!r}')
    modname = 'foo.bar'
    modpath = str(path)  # e.g. /tmp/foo/bar.py'
    mod = import_module_at_path(modname, modpath)
    assert isinstance(mod, types.ModuleType)
    assert mod.__name__ == modname
    assert mod.x == x


def test_import_module_at_path_package(tmp_path):
    path = tmp_path.joinpath('foo')
    path.mkdir(parents=True)
    init = path.joinpath('__init__.py')
    init.touch()
    x = 'hello'
    with init.open('w') as f:
        f.write(f'x={x!r}')
    modname = 'foo'
    modpath = str(path)
    mod = import_module_at_path(modname, modpath)
    assert isinstance(mod, types.ModuleType)
    assert mod.__name__ == modname
    assert mod.x == x


@pytest.mark.xfail
def test_import_module_at_path_bad_package_structure():
    raise NotImplementedError


def test_relpath_to_modname():
    relpath = 'ballet/util/_util.py'
    expected_modname = 'ballet.util._util'
    actual_modname = relpath_to_modname(relpath)
    assert actual_modname == expected_modname

    relpath = 'ballet/util/__init__.py'
    expected_modname = 'ballet.util'
    actual_modname = relpath_to_modname(relpath)
    assert actual_modname == expected_modname

    relpath = 'ballet/foo/bar/baz.zip'
    with pytest.raises(ValueError):
        relpath_to_modname(relpath)


def test_modname_to_relpath_module():
    modname = 'ballet.util._util'
    expected_relpath = 'ballet/util/_util.py'
    actual_relpath = modname_to_relpath(modname)
    assert actual_relpath == expected_relpath


def test_modname_to_relpath_package():
    modname = 'ballet.util'
    # mypackage.__file__ resolves to the __init__.py
    project_root = pathlib.Path(ballet.__file__).parent.parent

    expected_relpath = 'ballet/util/__init__.py'
    actual_relpath = modname_to_relpath(modname, project_root=project_root)
    assert actual_relpath == expected_relpath

    # TODO patch this
    # # without providing project root, behavior is undefined, as we don't
    # # know whether the relative path will resolve to a directory

    # # within a temporary directory, the relpath *should not* be a dir
    # with tempfile.TemporaryDirectory() as tmpdir:
    #     cwd = os.getcwd()
    #     try:
    #         os.chdir(tmpdir)
    #         actual_relpath = modname_to_relpath(modname)
    #         expected_relpath = 'ballet/util.py'
    #         self.assertEqual(actual_relpath, expected_relpath)
    #     finally:
    #         os.chdir(cwd)

    # # from the actual project root, the relpath *should* be a dir
    # cwd = os.getcwd()
    # try:
    #     os.chdir(str(project_root))
    #     actual_relpath = modname_to_relpath(modname)
    #     expected_relpath = 'ballet/util/__init__.py'
    #     self.assertEqual(actual_relpath, expected_relpath)
    # finally:
    #     os.chdir(cwd)


def test_modname_to_relpath_package_no_add_init():
    project_root = pathlib.Path(ballet.__file__).parent.parent
    # without add_init
    modname = 'ballet.util'
    add_init = False
    expected_relpath = 'ballet/util'
    actual_relpath = modname_to_relpath(
        modname, project_root=project_root, add_init=add_init)
    assert actual_relpath == expected_relpath



import os
import shlex
from subprocess import CalledProcessError, check_call
from textwrap import dedent
from types import ModuleType
from unittest.mock import patch

import numpy as np
import pandas as pd
import pytest

from ballet.compat import nullcontext
from ballet.eng.misc import IdentityTransformer
from ballet.feature import Feature
from ballet.pipeline import FeatureEngineeringPipeline
from ballet.project import FeatureEngineeringProject, make_feature_path
from ballet.templating import start_new_feature
from ballet.util.code import get_source
from ballet.util.git import make_commit_range, switch_to_new_branch
from ballet.util.log import logger
from tests.util import load_regression_data


def submit_feature(repo, contrib_dir, username, featurename, new_feature_str):
    feature_path = make_feature_path(contrib_dir, username, featurename)
    cc_kwargs = {
        'extra_context': {'username': username, 'featurename': featurename},
        'no_input': True,
    }
    result = start_new_feature(contrib_dir=contrib_dir, **cc_kwargs)
    with feature_path.open('w') as f:
        f.write(new_feature_str)

    added_files = [str(fn) for (fn, kind) in result if kind == 'file']
    repo.index.add(added_files)
    repo.index.commit(f'Add {feature_path} feature')


def make_feature_str(input):
    return dedent(
        f"""
        from ballet import Feature
        from ballet.eng.misc import IdentityTransformer
        input = {input!r}
        transformer = IdentityTransformer()
        feature = Feature(input, transformer)
        """
    ).strip()


@pytest.mark.slow
def test_validation_end_to_end(quickstart):
    project = quickstart.project
    slug = quickstart.package_slug
    base = project.path
    repo = quickstart.repo

    pkg = project.package
    assert isinstance(pkg, ModuleType)

    api = project.api
    assert isinstance(api, FeatureEngineeringProject)

    # no features at first
    features = api.features
    assert len(features) == 0

    # first providing a mock feature, call build
    mock_features = [Feature(input='A_1', transformer=IdentityTransformer())]
    with patch.object(api, 'collect', return_value=mock_features):
        X_df = pd.util.testing.makeCustomDataframe(5, 2)
        X_df.columns = ['A_0', 'A_1']
        result = api.engineer_features(X_df=X_df, y_df=[])
        assert np.shape(result.X) == (5, 1)
        assert isinstance(result.pipeline, FeatureEngineeringPipeline)

    # splice in a new version of foo.load_data.load_data
    # 1. 'src' needs to be hardcoded
    # 2. really bad - set load_data = load_regression_data which does not
    #    have the same args
    new_load_data_str = get_source(load_regression_data)
    p = base.joinpath('src', slug, 'load_data.py')
    with p.open('w') as f:
        f.write(new_load_data_str)
        f.write('\n')
        f.write('load_data=load_regression_data\n')

    # commit changes
    repo.index.add([str(p)])
    repo.index.commit('Load mock regression dataset')

    # call different validation routines
    def call_validate_all(ref=None):
        """Validate branch as if we were running on CI"""
        envvars = {
            'TRAVIS_BUILD_DIR': repo.working_tree_dir,
        }
        if ref is None:
            envvars['TRAVIS_PULL_REQUEST'] = 'false'
            envvars['TRAVIS_COMMIT_RANGE'] = make_commit_range(
                repo.commit('HEAD@{-1}').hexsha, repo.commit('HEAD').hexsha)
            envvars['TRAVIS_PULL_REQUEST_BRANCH'] = ''
            envvars['TRAVIS_BRANCH'] = repo.heads.master.name
        else:
            # TODO is this okay for testing?
            envvars['TRAVIS_PULL_REQUEST'] = str(1)
            envvars['TRAVIS_COMMIT_RANGE'] = make_commit_range(
                repo.heads.master.name,
                repo.commit(ref).hexsha)

        with patch.dict(os.environ, envvars):
            check_call(
                shlex.split('ballet validate -A'), cwd=base, env=os.environ)

    call_validate_all()

    # branch and write a new feature
    contrib_dir = base.joinpath('src', slug, 'features', 'contrib')
    ref = 'bob/feature-a'
    logger.info(f'Switching to branch {ref}, User Bob, Feature A')
    switch_to_new_branch(repo, ref)
    new_feature_str = make_feature_str('A_0')
    username = 'bob'
    featurename = 'A_0'
    submit_feature(repo, contrib_dir, username, featurename, new_feature_str)

    # call different validation routines
    logger.info('Validating User Bob, Feature A')
    call_validate_all(ref=ref)

    # merge branch with master
    logger.info('Merging into master')
    repo.git.checkout('master')
    repo.git.merge(ref, no_ff=True)

    # call different validation routines
    logger.info('Validating after merge')
    call_validate_all()

    # write another new feature
    ref = 'charlie/feature-z1'
    logger.info('Switching to branch ref, User Charlie, Feature Z_1')
    switch_to_new_branch(repo, ref)
    new_feature_str = make_feature_str('Z_1')
    username = 'charlie'
    featurename = 'Z_1'
    submit_feature(repo, contrib_dir, username, featurename, new_feature_str)

    # TODO we expect this feature to fail but it passes
    cm = pytest.raises(CalledProcessError) if False else nullcontext()
    with cm:
        logger.info('Validating User Charlie, Feature Z_1')
        call_validate_all(ref=ref)

    # write another new feature - redundancy
    ref = 'charlie/feature-a0'
    repo.git.checkout('master')
    switch_to_new_branch(repo, ref)
    new_feature_str = make_feature_str('A_0')
    username = 'charlie'
    featurename = 'A_0'
    submit_feature(repo, contrib_dir, username, featurename, new_feature_str)

    with pytest.raises(CalledProcessError):
        call_validate_all(ref=ref)



__all__ = []
try:
    from feature_engine.creation import MathematicalCombination
    __all__.append('MathematicalCombination')

    from feature_engine.creation import CombineWithReferenceFeature
    __all__.append('CombineWithReferenceFeature')

    from feature_engine.encoding import CountFrequencyEncoder
    __all__.append('CountFrequencyEncoder')

    from feature_engine.encoding import DecisionTreeEncoder
    __all__.append('DecisionTreeEncoder')

    from feature_engine.encoding import MeanEncoder
    __all__.append('MeanEncoder')

    from feature_engine.encoding import OneHotEncoder
    __all__.append('OneHotEncoder')

    from feature_engine.encoding import OrdinalEncoder
    __all__.append('OrdinalEncoder')

    from feature_engine.encoding import PRatioEncoder
    __all__.append('PRatioEncoder')

    from feature_engine.encoding import RareLabelEncoder
    __all__.append('RareLabelEncoder')

    from feature_engine.encoding import WoEEncoder
    __all__.append('WoEEncoder')

    from feature_engine.discretisation import ArbitraryDiscretiser
    __all__.append('ArbitraryDiscretiser')

    from feature_engine.discretisation import DecisionTreeDiscretiser
    __all__.append('DecisionTreeDiscretiser')

    from feature_engine.discretisation import EqualFrequencyDiscretiser
    __all__.append('EqualFrequencyDiscretiser')

    from feature_engine.discretisation import EqualWidthDiscretiser
    __all__.append('EqualWidthDiscretiser')

    from feature_engine.imputation import AddMissingIndicator
    __all__.append('AddMissingIndicator')

    from feature_engine.imputation import ArbitraryNumberImputer
    __all__.append('ArbitraryNumberImputer')

    from feature_engine.imputation import CategoricalImputer
    __all__.append('CategoricalImputer')

    from feature_engine.imputation import DropMissingData
    __all__.append('DropMissingData')

    from feature_engine.imputation import EndTailImputer
    __all__.append('EndTailImputer')

    from feature_engine.imputation import MeanMedianImputer
    __all__.append('MeanMedianImputer')

    from feature_engine.imputation import RandomSampleImputer
    __all__.append('RandomSampleImputer')

    from feature_engine.outliers import ArbitraryOutlierCapper
    __all__.append('ArbitraryOutlierCapper')

    from feature_engine.outliers import OutlierTrimmer
    __all__.append('OutlierTrimmer')

    from feature_engine.outliers import Winsorizer
    __all__.append('Winsorizer')

    from feature_engine.transformation import BoxCoxTransformer
    __all__.append('BoxCoxTransformer')

    from feature_engine.transformation import LogTransformer
    __all__.append('LogTransformer')

    from feature_engine.transformation import PowerTransformer
    __all__.append('PowerTransformer')

    from feature_engine.transformation import ReciprocalTransformer
    __all__.append('ReciprocalTransformer')

    from feature_engine.transformation import YeoJohnsonTransformer
    __all__.append('YeoJohnsonTransformer')
except ImportError:
    pass



from invoke import task


@task
def install(c, system=False):
    """install the package to the active Python's site-packages"""
    c.run('pip install -r requirements.txt')


def rmdir(c, name):
    c.run('rm -rf {name}')


def rm(c, pattern):
    c.run(f'find . -name \'{pattern}\' -exec rm -rf {{}} +')


@task
def clean(c):
    """Clean python artifacts"""
    for name in ['build', 'dist', '.eggs']:
        rmdir(c, name)

    for pattern in ['*.egg-info', '*.egg', '*.pyc', '*.pyo', '*~',
                    '__pycache__']:
        rm(c, pattern)



from setuptools import setup, find_packages

requirements = [
    'ballet==0.17.0',
]

setup(
    name='{{ cookiecutter.package_slug }}',
    version='0.1.0-dev',
    packages=find_packages(where='src', include=('{{ cookiecutter.package_slug }}', '{{ cookiecutter.package_slug }}.*')),
    package_dir={'': 'src'},
    install_requires=requirements,

    # metadata
    author='{{ cookiecutter.full_name.replace("\'", "\\\'") }}',
    author_email='{{ cookiecutter.email }}',
    description='A data science project built on the Ballet framework',
    url='https://github.com/{{ cookiecutter.github_owner }}/{{ cookiecutter.project_slug }}',
)



import sys


# sklearn compatibility
try:
    from sklearn.impute import SimpleImputer
except ImportError:
    import sklearn.preprocessing
    SimpleImputer = sklearn.preprocessing.Imputer


try:
    from os import PathLike
except ImportError:
    from pathlib import Path
    PathLike = (Path, )


# nullcontext new in 3.7?
try:
    from contextlib import nullcontext
except ImportError:
    from funcy import nullcontext


if sys.version_info >= (3, 7):
    import pathlib as _pathlib
    is_mount = _pathlib.Path.is_mount
else:
    # condensed python 3.8 pathlib source
    def is_mount(path):
        from pathlib import Path
        if not path.exists() or not path.is_dir():
            return False
        parent = Path(path.parent)
        try:
            parent_dev = parent.stat().st_dev
        except OSError:
            return False
        dev = path.stat().st_dev
        if dev != parent_dev:
            return True
        ino = path.stat().st_ino
        parent_ino = parent.stat().st_ino
        return ino == parent_ino



# -*- coding: utf-8 -*-

"""Top-level package for ballet."""

__author__ = 'Micah Smith'
__email__ = 'micahs@mit.edu'
__version__ = '0.17.0'

# filter warnings
import warnings  # noqa E402
warnings.filterwarnings(
    action='ignore', module='scipy', message='^internal gelsd')

# silence sklearn deprecation warnings
import logging  # noqa E402
logging.captureWarnings(True)
import sklearn  # noqa E402
logging.captureWarnings(False)
warnings.filterwarnings(
    action='ignore', module='sklearn', category=DeprecationWarning)
warnings.filterwarnings(
    action='ignore', module='sklearn', category=FutureWarning)

# configure module-level logging
from ballet.util.log import logger  # noqa E402
logger.addHandler(logging.NullHandler())

# re-export some names
from ballet.client import b  # noqa E402
from ballet.contrib import collect_contrib_features  # noqa E402
from ballet.feature import Feature  # noqa E402
from ballet.project import load_config, Project  # noqa E402

# for feature development, you really only need these two members
__all__ = (
    'b',
    'Feature',
)



import pathlib
import tempfile
from typing import List, Optional, Tuple

import funcy as fy
from cookiecutter.main import cookiecutter as _cookiecutter
from github import Github, GithubException

import ballet.util.git
from ballet.compat import PathLike
from ballet.exc import BalletError, ConfigurationError
from ballet.project import Project, detect_github_username
from ballet.util.fs import pwalk, synctree
from ballet.util.git import (
    DEFAULT_BRANCH, push_branches_to_remote, switch_to_new_branch,)
from ballet.util.log import logger
from ballet.util.typing import Pathy
from ballet.validation.project_structure.checks import (
    FEATURE_MODULE_NAME_REGEX, SUBPACKAGE_NAME_REGEX,)

TEMPLATES_PATH = pathlib.Path(__file__).resolve().parent.joinpath('templates')
FEATURE_TEMPLATE_PATH = TEMPLATES_PATH.joinpath('feature_template')
PROJECT_TEMPLATE_PATH = TEMPLATES_PATH.joinpath('project_template')


def _stringify_path(obj) -> str:
    return str(obj) if isinstance(obj, PathLike) else obj


@fy.wraps(_cookiecutter)
def cookiecutter(*args, **kwargs) -> str:
    """Call cookiecutter.main.cookiecutter after stringifying paths

    Return:
        project directory path
    """
    args = fy.walk(_stringify_path, args)
    kwargs = fy.walk_values(_stringify_path, kwargs)
    return _cookiecutter(*args, **kwargs)


def render_project_template(
    project_template_path: Optional[Pathy] = None,
    create_github_repo: bool = False,
    github_token: Optional[str] = None,
    **cc_kwargs
) -> str:
    """Generate a ballet project according to the project template

    If creating the GitHub repo is requested and the process fails for any
    reason, quickstart will complete successfully and users are instructed
    to read the corresponding section of the Maintainer's Guide to continue
    manually.

    Args:
        project_template_path: path to specific project template
        create_github_repo: whether to act to create the desired repo on
            GitHub after rendering the project. The repo will be owned by
            either the user or an org that the user has relevant permissions
            for, depending on what is entered during the quickstart prompts.
            If True, then a valid github token must also be provided.
        github_token: valid github token with appropriate permissions
        **cc_kwargs: options for the cookiecutter template
    """
    if project_template_path is None:
        project_template_path = PROJECT_TEMPLATE_PATH

    project_path = cookiecutter(project_template_path, **cc_kwargs)

    if create_github_repo:
        if github_token is None:
            raise ValueError('Need to provide github token')
        g = Github(github_token)

        # need to get params from new project config
        project = Project.from_path(project_path)
        owner = project.config.get('github.github_owner')
        name = project.config.get('project.project_slug')

        # create repo on github
        try:
            github_repo = ballet.util.git.create_github_repo(g, owner, name)
            logger.info(f'Created repo on GitHub at {github_repo.html_url}')
        except GithubException:
            logger.exception('Failed to create GitHub repo for this project')
            logger.warning(
                'Failed to create GitHub repo for this project...\n'
                'did you specify the intended repo owner, and do you have'
                ' permissions to create a repo under that owner?\n'
                'Try manually creating the repo: https://ballet.github.io/ballet/maintainer_guide.html#manual-repository-creation'  # noqa E501
            )
            return project_path

        # now push to remote
        # we don't need to set up the remote, as it has already been setup in
        # post_gen_hook.py
        local_repo = project.repo
        remote_name = project.config.get('github.remote')
        branches = [DEFAULT_BRANCH]
        try:
            push_branches_to_remote(local_repo, remote_name, branches)
        except BalletError:
            logger.exception('Failed to push branches to GitHub repo')
            logger.warning(
                'Failed to push branches to GitHub repo...\n'
                'Try manually pushing the branches: https://ballet.github.io/ballet/maintainer_guide.html#manual-repository-creation'  # noqa E501
            )
            return project_path

    return project_path


def render_feature_template(**cc_kwargs) -> str:
    """Create a stub for a new feature

    Args:
        **cc_kwargs: options for the cookiecutter template
    """
    feature_template_path = FEATURE_TEMPLATE_PATH
    return cookiecutter(feature_template_path, **cc_kwargs)


def _fail_if_feature_exists(dst: pathlib.Path) -> None:
    subpackage_name, feature_name = str(dst.parent), str(dst.name)
    if (
        dst.is_file()
        and fy.re_test(SUBPACKAGE_NAME_REGEX, subpackage_name)
        and fy.re_test(FEATURE_MODULE_NAME_REGEX, feature_name)
    ):
        raise FileExistsError(f'The feature already exists here: {dst}')


def start_new_feature(
    contrib_dir: Pathy = None,
    branching: bool = True,
    **cc_kwargs
) -> List[Tuple[pathlib.Path, str]]:
    """Start a new feature within a ballet project

    If run from default branch, by default will attempt to switch to a new
    branch for this feature, given by `<username>/feature-<featurename>`. By
    default, will prompt the user for input using cookiecutter's input
    interface.

    Renders the feature template into a temporary directory, then copies the
    feature files into the proper path within the contrib directory.

    Args:
        contrib_dir: directory under which to place contributed features
        branching: whether to attempt to manage branching
        **cc_kwargs: options for the cookiecutter template

    Raises:
        ballet.exc.BalletError: the new feature has the same name as an
            existing one
    """
    if contrib_dir is not None:
        try:
            project = Project.from_path(contrib_dir, ascend=True)
            default_username = detect_github_username(project)
        except ConfigurationError:
            default_username = 'username'
    else:
        project = Project.from_cwd()
        contrib_dir = project.config.get('contrib.module_path')
        default_username = detect_github_username(project)

    # inject default username into context
    cc_kwargs.setdefault('extra_context', {})
    cc_kwargs['extra_context'].update({'_default_username': default_username})

    with tempfile.TemporaryDirectory() as tempdir:
        # render feature template
        output_dir = tempdir
        cc_kwargs['output_dir'] = output_dir
        rendered_dir = render_feature_template(**cc_kwargs)

        # clean pyc files from rendered dir
        for path in pwalk(rendered_dir, topdown=False):
            if path.suffix == '.pyc':
                path.unlink()
            if path.name == '__pycache__':
                with fy.suppress(OSError):
                    path.rmdir()

        # copy into contrib dir
        src = rendered_dir
        dst = contrib_dir
        result = synctree(src, dst, onexist=_fail_if_feature_exists)

    target_branch = None
    if branching and project.on_master:

        # try to set the target branch name
        paths = [path for path, kind in result if kind == 'file']
        for path in paths:
            parts = pathlib.Path(path).parts
            subpackage, module = parts[-2], parts[-1]
            user_match = fy.re_find(SUBPACKAGE_NAME_REGEX, subpackage)
            feature_match = fy.re_find(FEATURE_MODULE_NAME_REGEX, module)
            if feature_match:
                username = user_match['username']
                featurename = feature_match['featurename'].replace('_', '-')
                target_branch = f'{username}/feature-{featurename}'

        if target_branch is not None:
            switch_to_new_branch(project.repo, target_branch)

    _log_start_new_feature_success(result)
    _log_switch_to_new_branch(target_branch)

    return result


def _log_start_new_feature_success(result: List[Tuple[pathlib.Path, str]]):
    logger.info('Start new feature successful')
    for (name, kind) in result:
        if kind == 'file' and '__init__' not in str(name):
            relname = pathlib.Path(name).relative_to(pathlib.Path.cwd())
            logger.info(f'Created {relname}')


def _log_switch_to_new_branch(branch: Optional[str]):
    if branch is not None:
        logger.info(f'Switched to branch {branch}')
